import os
import glob
import json
import zipfile
import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict, deque
from typing import Union, List, Callable, Optional

import gym
import cloudpickle
import numpy as np
import tensorflow as tf

from neorl.rl.baselines.shared.misc_util import set_global_seeds
from neorl.rl.baselines.shared.save_util import data_to_json, json_to_data, params_to_bytes, bytes_to_params
from neorl.rl.baselines.shared.policies import get_policy_from_name, ActorCriticPolicy
from neorl.rl.baselines.shared.runners import AbstractEnvRunner
from neorl.rl.baselines.shared.vec_env import (VecEnvWrapper, VecEnv, DummyVecEnv,
                                             VecNormalize, unwrap_vec_normalize)
from neorl.rl.baselines.shared.callbacks import BaseCallback, CallbackList, ConvertCallback
from neorl.rl.baselines.shared import logger


class BaseRLModel(ABC):
    """
    The base RL model

    :param policy: (BasePolicy) Policy object
    :param env: (Gym environment) The environment to learn from
                (if registered in Gym, can be str. Can be None for loading trained models)
    :param verbose: (int) the verbosity level: 0 none, 1 training information, 2 tensorflow debug
    :param requires_vec_env: (bool) Does this model require a vectorized environment
    :param policy_base: (BasePolicy) the base policy used by this method
    :param policy_kwargs: (dict) additional arguments to be passed to the policy on creation
    :param seed: (int) Seed for the pseudo-random generators (python, numpy, tensorflow).
        If None (default), use random seed. Note that if you want completely deterministic
        results, you must set `n_cpu_tf_sess` to 1.
    :param n_cpu_tf_sess: (int) The number of threads for TensorFlow operations
        If None, the number of cpu of the current machine will be used.
    """

    def __init__(self, policy, env, verbose=0, *, requires_vec_env, policy_base,
                 policy_kwargs=None, seed=None, n_cpu_tf_sess=None):
        if isinstance(policy, str) and policy_base is not None:
            self.policy = get_policy_from_name(policy_base, policy)
        else:
            self.policy = policy
        self.env = env
        self.verbose = verbose
        self._requires_vec_env = requires_vec_env
        self.policy_kwargs = {} if policy_kwargs is None else policy_kwargs
        self.observation_space = None
        self.action_space = None
        self.n_envs = None
        self._vectorize_action = False
        self.num_timesteps = 0
        self.graph = None
        self.sess = None
        self.params = None
        self.seed = seed
        self._param_load_ops = None
        self.n_cpu_tf_sess = n_cpu_tf_sess
        self.episode_reward = None
        self.ep_info_buf = None

        if env is not None:
            if isinstance(env, str):
                if self.verbose >= 1:
                    print("Creating environment from the given name, wrapped in a DummyVecEnv.")
                self.env = env = DummyVecEnv([lambda: gym.make(env)])

            self.observation_space = env.observation_space
            self.action_space = env.action_space
            if requires_vec_env:
                if isinstance(env, VecEnv):
                    self.n_envs = env.num_envs
                else:
                    # The model requires a VecEnv
                    # wrap it in a DummyVecEnv to avoid error
                    self.env = DummyVecEnv([lambda: env])
                    if self.verbose >= 1:
                        print("Wrapping the env in a DummyVecEnv.")
                    self.n_envs = 1
            else:
                if isinstance(env, VecEnv):
                    if env.num_envs == 1:
                        self.env = _UnvecWrapper(env)
                        self._vectorize_action = True
                    else:
                        raise ValueError("Error: the model requires a non vectorized environment or a single vectorized"
                                         " environment.")
                self.n_envs = 1

        # Get VecNormalize object if it exists
        self._vec_normalize_env = unwrap_vec_normalize(self.env)

    def get_env(self):
#        """
#        returns the current environment (can be None if not defined)
#
#        :return: (Gym Environment) The current environment
#        """
        return self.env

    def get_vec_normalize_env(self) -> Optional[VecNormalize]:
#        """
#        Return the ``VecNormalize`` wrapper of the training env
#        if it exists.
#
#        :return: Optional[VecNormalize] The ``VecNormalize`` env.
#        """
        return self._vec_normalize_env

    def set_env(self, env):
#        """
#        Checks the validity of the environment, and if it is coherent, set it as the current environment.
#
#        :param env: (Gym Environment) The environment for learning a policy
#        """
        if env is None and self.env is None:
            if self.verbose >= 1:
                print("Loading a model without an environment, "
                      "this model cannot be trained until it has a valid environment.")
            return
        elif env is None:
            raise ValueError("Error: trying to replace the current environment with None")

        # sanity checking the environment
        assert self.observation_space == env.observation_space, \
            "Error: the environment passed must have at least the same observation space as the model was trained on."
        assert self.action_space == env.action_space, \
            "Error: the environment passed must have at least the same action space as the model was trained on."
        if self._requires_vec_env:
            assert isinstance(env, VecEnv), \
                "Error: the environment passed is not a vectorized environment, however {} requires it".format(
                    self.__class__.__name__)
            assert not self.policy.recurrent or self.n_envs == env.num_envs, \
                "Error: the environment passed must have the same number of environments as the model was trained on." \
                "This is due to the Lstm policy not being capable of changing the number of environments."
            self.n_envs = env.num_envs
        else:
            # for models that dont want vectorized environment, check if they make sense and adapt them.
            # Otherwise tell the user about this issue
            if isinstance(env, VecEnv):
                if env.num_envs == 1:
                    env = _UnvecWrapper(env)
                    self._vectorize_action = True
                else:
                    raise ValueError("Error: the model requires a non vectorized environment or a single vectorized "
                                     "environment.")
            else:
                self._vectorize_action = False

            self.n_envs = 1

        self.env = env
        self._vec_normalize_env = unwrap_vec_normalize(env)

        # Invalidated by environment change.
        self.episode_reward = None
        self.ep_info_buf = None

    def _init_num_timesteps(self, reset_num_timesteps=True):
#        """
#        Initialize and resets num_timesteps (total timesteps since beginning of training)
#        if needed. Mainly used logging and plotting (tensorboard).
#
#        :param reset_num_timesteps: (bool) Set it to false when continuing training
#            to not create new plotting curves in tensorboard.
#        :return: (bool) Whether a new tensorboard log needs to be created
#        """
        if reset_num_timesteps:
            self.num_timesteps = 0

        new_tb_log = self.num_timesteps == 0
        return new_tb_log

    @abstractmethod
    def setup_model(self):
#        """
#        Create all the functions and tensorflow graphs necessary to train the model
#        """
        pass

    def _init_callback(self,
                      callback: Union[None, Callable, List[BaseCallback], BaseCallback]
                      ) -> BaseCallback:
#        """
#        :param callback: (Union[None, Callable, List[BaseCallback], BaseCallback])
#        :return: (BaseCallback)
#        """
        # Convert a list of callbacks into a callback
        if isinstance(callback, list):
            callback = CallbackList(callback)
        # Convert functional callback to object
        if not isinstance(callback, BaseCallback):
            callback = ConvertCallback(callback)

        callback.init_callback(self)
        return callback

    def set_random_seed(self, seed: Optional[int]) -> None:
#        """
#        :param seed: (Optional[int]) Seed for the pseudo-random generators. If None,
#            do not change the seeds.
#        """
        # Ignore if the seed is None
        if seed is None:
            return
        # Seed python, numpy and tf random generator
        set_global_seeds(seed)
        if self.env is not None:
            self.env.seed(seed)
            # Seed the action space
            # useful when selecting random actions
            self.env.action_space.seed(seed)
        self.action_space.seed(seed)

    def _setup_learn(self):
#        """
#        Check the environment.
#        """
        if self.env is None:
            raise ValueError("Error: cannot train the model without a valid environment, please set an environment with"
                             "set_env(self, env) method.")
        if self.episode_reward is None:
            self.episode_reward = np.zeros((self.n_envs,))
        if self.ep_info_buf is None:
            self.ep_info_buf = deque(maxlen=100)

    @abstractmethod
    def get_parameter_list(self):
#        """
#        Get tensorflow Variables of model's parameters
#
#        This includes all variables necessary for continuing training (saving / loading).
#
#        :return: (list) List of tensorflow Variables
#        """
        pass

    def get_parameters(self):
#        """
#        Get current model parameters as dictionary of variable name -> ndarray.
#
#        :return: (OrderedDict) Dictionary of variable name -> ndarray of model's parameters.
#        """
        parameters = self.get_parameter_list()
        parameter_values = self.sess.run(parameters)
        return_dictionary = OrderedDict((param.name, value) for param, value in zip(parameters, parameter_values))
        return return_dictionary

    def _setup_load_operations(self):
#        """
#        Create tensorflow operations for loading model parameters
#        """
        # Assume tensorflow graphs are static -> check
        # that we only call this function once
        if self._param_load_ops is not None:
            raise RuntimeError("Parameter load operations have already been created")
        # For each loadable parameter, create appropiate
        # placeholder and an assign op, and store them to
        # self.load_param_ops as dict of variable.name -> (placeholder, assign)
        loadable_parameters = self.get_parameter_list()
        # Use OrderedDict to store order for backwards compatibility with
        # list-based params
        self._param_load_ops = OrderedDict()
        with self.graph.as_default():
            for param in loadable_parameters:
                placeholder = tf.placeholder(dtype=param.dtype, shape=param.shape)
                # param.name is unique (tensorflow variables have unique names)
                self._param_load_ops[param.name] = (placeholder, param.assign(placeholder))

    @abstractmethod
    def _get_pretrain_placeholders(self):
#        """
#        Return the placeholders needed for the pretraining:
#        - obs_ph: observation placeholder
#        - actions_ph will be population with an action from the environment
#            (from the expert dataset)
#        - deterministic_actions_ph: e.g., in the case of a Gaussian policy,
#            the mean.
#
#        :return: ((tf.placeholder)) (obs_ph, actions_ph, deterministic_actions_ph)
#        """
        pass

    def pretrain(self, dataset, n_epochs=10, learning_rate=1e-4,
                 adam_epsilon=1e-8, val_interval=None):
#        """
#        Pretrain a model using behavior cloning:
#        supervised learning given an expert dataset.
#
#        NOTE: only Box and Discrete spaces are supported for now.
#
#        :param dataset: (ExpertDataset) Dataset manager
#        :param n_epochs: (int) Number of iterations on the training set
#        :param learning_rate: (float) Learning rate
#        :param adam_epsilon: (float) the epsilon value for the adam optimizer
#        :param val_interval: (int) Report training and validation losses every n epochs.
#            By default, every 10th of the maximum number of epochs.
#        :return: (BaseRLModel) the pretrained model
#        """
        continuous_actions = isinstance(self.action_space, gym.spaces.Box)
        discrete_actions = isinstance(self.action_space, gym.spaces.Discrete)

        assert discrete_actions or continuous_actions, 'Only Discrete and Box action spaces are supported'

        # Validate the model every 10% of the total number of iteration
        if val_interval is None:
            # Prevent modulo by zero
            if n_epochs < 10:
                val_interval = 1
            else:
                val_interval = int(n_epochs / 10)

        with self.graph.as_default():
            with tf.variable_scope('pretrain'):
                if continuous_actions:
                    obs_ph, actions_ph, deterministic_actions_ph = self._get_pretrain_placeholders()
                    loss = tf.reduce_mean(tf.square(actions_ph - deterministic_actions_ph))
                else:
                    obs_ph, actions_ph, actions_logits_ph = self._get_pretrain_placeholders()
                    # actions_ph has a shape if (n_batch,), we reshape it to (n_batch, 1)
                    # so no additional changes is needed in the dataloader
                    actions_ph = tf.expand_dims(actions_ph, axis=1)
                    one_hot_actions = tf.one_hot(actions_ph, self.action_space.n)
                    loss = tf.nn.softmax_cross_entropy_with_logits_v2(
                        logits=actions_logits_ph,
                        labels=tf.stop_gradient(one_hot_actions)
                    )
                    loss = tf.reduce_mean(loss)
                optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate, epsilon=adam_epsilon)
                optim_op = optimizer.minimize(loss, var_list=self.params)

            self.sess.run(tf.global_variables_initializer())

        if self.verbose > 0:
            print("Pretraining with Behavior Cloning...")

        for epoch_idx in range(int(n_epochs)):
            train_loss = 0.0
            # Full pass on the training set
            for _ in range(len(dataset.train_loader)):
                expert_obs, expert_actions = dataset.get_next_batch('train')
                feed_dict = {
                    obs_ph: expert_obs,
                    actions_ph: expert_actions,
                }
                train_loss_, _ = self.sess.run([loss, optim_op], feed_dict)
                train_loss += train_loss_

            train_loss /= len(dataset.train_loader)

            if self.verbose > 0 and (epoch_idx + 1) % val_interval == 0:
                val_loss = 0.0
                # Full pass on the validation set
                for _ in range(len(dataset.val_loader)):
                    expert_obs, expert_actions = dataset.get_next_batch('val')
                    val_loss_, = self.sess.run([loss], {obs_ph: expert_obs,
                                                        actions_ph: expert_actions})
                    val_loss += val_loss_

                val_loss /= len(dataset.val_loader)
                if self.verbose > 0:
                    print("==== Training progress {:.2f}% ====".format(100 * (epoch_idx + 1) / n_epochs))
                    print('Epoch {}'.format(epoch_idx + 1))
                    print("Training loss: {:.6f}, Validation loss: {:.6f}".format(train_loss, val_loss))
                    print()
            # Free memory
            del expert_obs, expert_actions
        if self.verbose > 0:
            print("Pretraining done.")
        return self

    @abstractmethod
    def learn(self, total_timesteps, callback=None, log_interval=100, tb_log_name="run",
              reset_num_timesteps=True):
        """
        Return a trained model.

        :param total_timesteps: (int) The total number of samples to train on
        :param callback: (Union[callable, [callable], BaseCallback])
            function called at every steps with state of the algorithm.
            It takes the local and global variables. If it returns False, training is aborted.
            When the callback inherits from BaseCallback, you will have access
            to additional stages of the training (training start/end),
            please read the documentation for more details.
        :param log_interval: (int) The number of timesteps before logging.
        :param tb_log_name: (str) the name of the run for tensorboard log
        :param reset_num_timesteps: (bool) whether or not to reset the current timestep number (used in logging)
        :return: (BaseRLModel) the trained model
        """
        pass

    @abstractmethod
    def predict(self, observation, state=None, mask=None, deterministic=False):
        """
        Get the model's action from an observation

        :param observation: (np.ndarray) the input observation
        :param state: (np.ndarray) The last states (can be None, used in recurrent policies)
        :param mask: (np.ndarray) The last masks (can be None, used in recurrent policies)
        :param deterministic: (bool) Whether or not to return deterministic actions.
        :return: (np.ndarray, np.ndarray) the model's action and the next state (used in recurrent policies)
        """
        pass

    @abstractmethod
    def action_probability(self, observation, state=None, mask=None, actions=None, logp=False):
#        """
#        If ``actions`` is ``None``, then get the model's action probability distribution from a given observation.
#
#        Depending on the action space the output is:
#            - Discrete: probability for each possible action
#            - Box: mean and standard deviation of the action output
#
#        However if ``actions`` is not ``None``, this function will return the probability that the given actions are
#        taken with the given parameters (observation, state, ...) on this model. For discrete action spaces, it
#        returns the probability mass; for continuous action spaces, the probability density. This is since the
#        probability mass will always be zero in continuous spaces, see http://blog.christianperone.com/2019/01/
#        for a good explanation
#
#        :param observation: (np.ndarray) the input observation
#        :param state: (np.ndarray) The last states (can be None, used in recurrent policies)
#        :param mask: (np.ndarray) The last masks (can be None, used in recurrent policies)
#        :param actions: (np.ndarray) (OPTIONAL) For calculating the likelihood that the given actions are chosen by
#            the model for each of the given parameters. Must have the same number of actions and observations.
#            (set to None to return the complete action probability distribution)
#        :param logp: (bool) (OPTIONAL) When specified with actions, returns probability in log-space.
#            This has no effect if actions is None.
#        :return: (np.ndarray) the model's (log) action probability
#        """
        pass

    def load_parameters(self, load_path_or_dict, exact_match=True):
#        """
#        Load model parameters from a file or a dictionary
#
#        Dictionary keys should be tensorflow variable names, which can be obtained
#        with ``get_parameters`` function. If ``exact_match`` is True, dictionary
#        should contain keys for all model's parameters, otherwise RunTimeError
#        is raised. If False, only variables included in the dictionary will be updated.
#
#        This does not load agent's hyper-parameters.
#
#        .. warning::
#            This function does not update trainer/optimizer variables (e.g. momentum).
#            As such training after using this function may lead to less-than-optimal results.
#
#        :param load_path_or_dict: (str or file-like or dict) Save parameter location
#            or dict of parameters as variable.name -> ndarrays to be loaded.
#        :param exact_match: (bool) If True, expects load dictionary to contain keys for
#            all variables in the model. If False, loads parameters only for variables
#            mentioned in the dictionary. Defaults to True.
#        """
        # Make sure we have assign ops
        if self._param_load_ops is None:
            self._setup_load_operations()

        if isinstance(load_path_or_dict, dict):
            # Assume `load_path_or_dict` is dict of variable.name -> ndarrays we want to load
            params = load_path_or_dict
        elif isinstance(load_path_or_dict, list):
            warnings.warn("Loading model parameters from a list. This has been replaced " +
                          "with parameter dictionaries with variable names and parameters. " +
                          "If you are loading from a file, consider re-saving the file.",
                          DeprecationWarning)
            # Assume `load_path_or_dict` is list of ndarrays.
            # Create param dictionary assuming the parameters are in same order
            # as `get_parameter_list` returns them.
            params = dict()
            for i, param_name in enumerate(self._param_load_ops.keys()):
                params[param_name] = load_path_or_dict[i]
        else:
            # Assume a filepath or file-like.
            # Use existing deserializer to load the parameters.
            # We only need the parameters part of the file, so
            # only load that part.
            _, params = BaseRLModel._load_from_file(load_path_or_dict, load_data=False)
            params = dict(params)

        feed_dict = {}
        param_update_ops = []
        # Keep track of not-updated variables
        not_updated_variables = set(self._param_load_ops.keys())
        for param_name, param_value in params.items():
            placeholder, assign_op = self._param_load_ops[param_name]
            feed_dict[placeholder] = param_value
            # Create list of tf.assign operations for sess.run
            param_update_ops.append(assign_op)
            # Keep track which variables are updated
            not_updated_variables.remove(param_name)

        # Check that we updated all parameters if exact_match=True
        if exact_match and len(not_updated_variables) > 0:
            raise RuntimeError("Load dictionary did not contain all variables. " +
                               "Missing variables: {}".format(", ".join(not_updated_variables)))

        self.sess.run(param_update_ops, feed_dict=feed_dict)

    @abstractmethod
    def save(self, save_path, cloudpickle=False):
        """
        Save the current parameters to file

        :param save_path: (str or file-like) The save location
        :param cloudpickle: (bool) Use older cloudpickle format instead of zip-archives.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def load(cls, load_path, env=None, custom_objects=None, **kwargs):
#        """
#        Load the model from file
#
#        :param load_path: (str or file-like) the saved parameter location
#        :param env: (Gym Environment) the new environment to run the loaded model on
#            (can be None if you only need prediction from a trained model)
#        :param custom_objects: (dict) Dictionary of objects to replace
#            upon loading. If a variable is present in this dictionary as a
#            key, it will not be deserialized and the corresponding item
#            will be used instead. Similar to custom_objects in
#            `keras.models.load_model`. Useful when you have an object in
#            file that can not be deserialized.
#        :param kwargs: extra arguments to change the model when loading
#        """
        raise NotImplementedError()

    @staticmethod
    def _save_to_file_cloudpickle(save_path, data=None, params=None):
#        """Legacy code for saving models with cloudpickle
#
#        :param save_path: (str or file-like) Where to store the model
#        :param data: (OrderedDict) Class parameters being stored
#        :param params: (OrderedDict) Model parameters being stored
#        """
        if isinstance(save_path, str):
            _, ext = os.path.splitext(save_path)
            if ext == "":
                save_path += ".pkl"

            with open(save_path, "wb") as file_:
                cloudpickle.dump((data, params), file_)
        else:
            # Here save_path is a file-like object, not a path
            cloudpickle.dump((data, params), save_path)

    @staticmethod
    def _save_to_file_zip(save_path, data=None, params=None):
#        """Save model to a .zip archive
#
#        :param save_path: (str or file-like) Where to store the model
#        :param data: (OrderedDict) Class parameters being stored
#        :param params: (OrderedDict) Model parameters being stored
#        """
        # data/params can be None, so do not
        # try to serialize them blindly
        if data is not None:
            serialized_data = data_to_json(data)
        if params is not None:
            serialized_params = params_to_bytes(params)
            # We also have to store list of the parameters
            # to store the ordering for OrderedDict.
            # We can trust these to be strings as they
            # are taken from the Tensorflow graph.
            serialized_param_list = json.dumps(
                list(params.keys()),
                indent=4
            )

        # Check postfix if save_path is a string
        if isinstance(save_path, str):
            _, ext = os.path.splitext(save_path)
            if ext == "":
                save_path += ".zip"

        # Create a zip-archive and write our objects
        # there. This works when save_path
        # is either str or a file-like
        with zipfile.ZipFile(save_path, "w") as file_:
            # Do not try to save "None" elements
            if data is not None:
                file_.writestr("data", serialized_data)
            if params is not None:
                file_.writestr("parameters", serialized_params)
                file_.writestr("parameter_list", serialized_param_list)

    @staticmethod
    def _save_to_file(save_path, data=None, params=None, cloudpickle=False):
#        """Save model to a zip archive or cloudpickle file.
#
#        :param save_path: (str or file-like) Where to store the model
#        :param data: (OrderedDict) Class parameters being stored
#        :param params: (OrderedDict) Model parameters being stored
#        :param cloudpickle: (bool) Use old cloudpickle format
#            (stable-baselines<=2.7.0) instead of a zip archive.
#        """
        if cloudpickle:
            BaseRLModel._save_to_file_cloudpickle(save_path, data, params)
        else:
            BaseRLModel._save_to_file_zip(save_path, data, params)

    @staticmethod
    def _load_from_file_cloudpickle(load_path):
        """Legacy code for loading older models stored with cloudpickle

        :param load_path: (str or file-like) where from to load the file
        :return: (dict, OrderedDict) Class parameters and model parameters
        """
        if isinstance(load_path, str):
            if not os.path.exists(load_path):
                if os.path.exists(load_path + ".pkl"):
                    load_path += ".pkl"
                else:
                    raise ValueError("Error: the file {} could not be found".format(load_path))

            with open(load_path, "rb") as file_:
                data, params = cloudpickle.load(file_)
        else:
            # Here load_path is a file-like object, not a path
            data, params = cloudpickle.load(load_path)

        return data, params

    @staticmethod
    def _load_from_file(load_path, load_data=True, custom_objects=None):
        """Load model data from a .zip archive

        :param load_path: (str or file-like) Where to load model from
        :param load_data: (bool) Whether we should load and return data
            (class parameters). Mainly used by `load_parameters` to
            only load model parameters (weights).
        :param custom_objects: (dict) Dictionary of objects to replace
            upon loading. If a variable is present in this dictionary as a
            key, it will not be deserialized and the corresponding item
            will be used instead. Similar to custom_objects in
            `keras.models.load_model`. Useful when you have an object in
            file that can not be deserialized.
        :return: (dict, OrderedDict) Class parameters and model parameters
        """
        # Check if file exists if load_path is
        # a string
        if isinstance(load_path, str):
            if not os.path.exists(load_path):
                if os.path.exists(load_path + ".zip"):
                    load_path += ".zip"
                else:
                    raise ValueError("Error: the file {} could not be found".format(load_path))

        # Open the zip archive and load data.
        try:
            with zipfile.ZipFile(load_path, "r") as file_:
                namelist = file_.namelist()
                # If data or parameters is not in the
                # zip archive, assume they were stored
                # as None (_save_to_file allows this).
                data = None
                params = None
                if "data" in namelist and load_data:
                    # Load class parameters and convert to string
                    # (Required for json library in Python 3.5)
                    json_data = file_.read("data").decode()
                    data = json_to_data(json_data, custom_objects=custom_objects)

                if "parameters" in namelist:
                    # Load parameter list and and parameters
                    parameter_list_json = file_.read("parameter_list").decode()
                    parameter_list = json.loads(parameter_list_json)
                    serialized_params = file_.read("parameters")
                    params = bytes_to_params(
                        serialized_params, parameter_list
                    )
        except zipfile.BadZipFile:
            # load_path wasn't a zip file. Possibly a cloudpickle
            # file. Show a warning and fall back to loading cloudpickle.
            warnings.warn("It appears you are loading from a file with old format. " +
                          "Older cloudpickle format has been replaced with zip-archived " +
                          "models. Consider saving the model with new format.",
                          DeprecationWarning)
            # Attempt loading with the cloudpickle format.
            # If load_path is file-like, seek back to beginning of file
            if not isinstance(load_path, str):
                load_path.seek(0)
            data, params = BaseRLModel._load_from_file_cloudpickle(load_path)

        return data, params

    @staticmethod
    def _softmax(x_input):
        """
        An implementation of softmax.

        :param x_input: (numpy float) input vector
        :return: (numpy float) output vector
        """
        x_exp = np.exp(x_input.T - np.max(x_input.T, axis=0))
        return (x_exp / x_exp.sum(axis=0)).T

    @staticmethod
    def _is_vectorized_observation(observation, observation_space):
        """
        For every observation type, detects and validates the shape,
        then returns whether or not the observation is vectorized.

        :param observation: (np.ndarray) the input observation to validate
        :param observation_space: (gym.spaces) the observation space
        :return: (bool) whether the given observation is vectorized or not
        """
        if isinstance(observation_space, gym.spaces.Box):
            if observation.shape == observation_space.shape:
                return False
            elif observation.shape[1:] == observation_space.shape:
                return True
            else:
                raise ValueError("Error: Unexpected observation shape {} for ".format(observation.shape) +
                                 "Box environment, please use {} ".format(observation_space.shape) +
                                 "or (n_env, {}) for the observation shape."
                                 .format(", ".join(map(str, observation_space.shape))))
        elif isinstance(observation_space, gym.spaces.Discrete):
            if observation.shape == ():  # A numpy array of a number, has shape empty tuple '()'
                return False
            elif len(observation.shape) == 1:
                return True
            else:
                raise ValueError("Error: Unexpected observation shape {} for ".format(observation.shape) +
                                 "Discrete environment, please use (1,) or (n_env, 1) for the observation shape.")
        elif isinstance(observation_space, gym.spaces.MultiDiscrete):
            if observation.shape == (len(observation_space.nvec),):
                return False
            elif len(observation.shape) == 2 and observation.shape[1] == len(observation_space.nvec):
                return True
            else:
                raise ValueError("Error: Unexpected observation shape {} for MultiDiscrete ".format(observation.shape) +
                                 "environment, please use ({},) or ".format(len(observation_space.nvec)) +
                                 "(n_env, {}) for the observation shape.".format(len(observation_space.nvec)))
        elif isinstance(observation_space, gym.spaces.MultiBinary):
            if observation.shape == (observation_space.n,):
                return False
            elif len(observation.shape) == 2 and observation.shape[1] == observation_space.n:
                return True
            else:
                raise ValueError("Error: Unexpected observation shape {} for MultiBinary ".format(observation.shape) +
                                 "environment, please use ({},) or ".format(observation_space.n) +
                                 "(n_env, {}) for the observation shape.".format(observation_space.n))
        else:
            raise ValueError("Error: Cannot determine if the observation is vectorized with the space type {}."
                             .format(observation_space))


class ActorCriticRLModel(BaseRLModel):
    """
    The base class for Actor critic model

    :param policy: (BasePolicy) Policy object
    :param env: (Gym environment) The environment to learn from
                (if registered in Gym, can be str. Can be None for loading trained models)
    :param verbose: (int) the verbosity level: 0 none, 1 training information, 2 tensorflow debug
    :param policy_base: (BasePolicy) the base policy used by this method (default=ActorCriticPolicy)
    :param requires_vec_env: (bool) Does this model require a vectorized environment
    :param policy_kwargs: (dict) additional arguments to be passed to the policy on creation
    :param seed: (int) Seed for the pseudo-random generators (python, numpy, tensorflow).
        If None (default), use random seed. Note that if you want completely deterministic
        results, you must set `n_cpu_tf_sess` to 1.
    :param n_cpu_tf_sess: (int) The number of threads for TensorFlow operations
        If None, the number of cpu of the current machine will be used.
    """

    def __init__(self, policy, env, _init_setup_model=False, verbose=0, policy_base=ActorCriticPolicy,
                 requires_vec_env=False, policy_kwargs=None, seed=None, n_cpu_tf_sess=None):
        super(ActorCriticRLModel, self).__init__(policy, env, verbose=verbose, requires_vec_env=requires_vec_env,
                                                 policy_base=policy_base, policy_kwargs=policy_kwargs,
                                                 seed=seed, n_cpu_tf_sess=n_cpu_tf_sess)

        self.sess = None
        self.initial_state = None
        self.step = None
        self.proba_step = None
        self.params = None
        self._runner = None

    def _make_runner(self) -> AbstractEnvRunner:
        """Builds a new Runner.

        Lazily called whenever `self.runner` is accessed and `self._runner is None`.
        """
        raise NotImplementedError("This model is not configured to use a Runner")

    @property
    def runner(self) -> AbstractEnvRunner:
        if self._runner is None:
            self._runner = self._make_runner()
        return self._runner

    def set_env(self, env):
        self._runner = None  # New environment invalidates `self._runner`.
        super().set_env(env)

    @abstractmethod
    def setup_model(self):
        pass

    @abstractmethod
    def learn(self, total_timesteps, callback=None,
              log_interval=100, tb_log_name="run", reset_num_timesteps=True):
        pass

    def predict(self, observation, state=None, mask=None, deterministic=False):
        if state is None:
            state = self.initial_state
        if mask is None:
            mask = [False for _ in range(self.n_envs)]
        observation = np.array(observation)
        vectorized_env = self._is_vectorized_observation(observation, self.observation_space)

        observation = observation.reshape((-1,) + self.observation_space.shape)
        actions, _, states, _ = self.step(observation, state, mask, deterministic=deterministic)

        clipped_actions = actions
        # Clip the actions to avoid out of bound error
        if isinstance(self.action_space, gym.spaces.Box):
            clipped_actions = np.clip(actions, self.action_space.low, self.action_space.high)

        if not vectorized_env:
            if state is not None:
                raise ValueError("Error: The environment must be vectorized when using recurrent policies.")
            clipped_actions = clipped_actions[0]

        return clipped_actions, states

    def action_probability(self, observation, state=None, mask=None, actions=None, logp=False):
        if state is None:
            state = self.initial_state
        if mask is None:
            mask = [False for _ in range(self.n_envs)]
        observation = np.array(observation)
        vectorized_env = self._is_vectorized_observation(observation, self.observation_space)

        observation = observation.reshape((-1,) + self.observation_space.shape)
        actions_proba = self.proba_step(observation, state, mask)

        if len(actions_proba) == 0:  # empty list means not implemented
            warnings.warn("Warning: action probability is not implemented for {} action space. Returning None."
                          .format(type(self.action_space).__name__))
            return None

        if actions is not None:  # comparing the action distribution, to given actions
            prob = None
            logprob = None
            actions = np.array([actions])
            if isinstance(self.action_space, gym.spaces.Discrete):
                actions = actions.reshape((-1,))
                assert observation.shape[0] == actions.shape[0], \
                    "Error: batch sizes differ for actions and observations."
                prob = actions_proba[np.arange(actions.shape[0]), actions]

            elif isinstance(self.action_space, gym.spaces.MultiDiscrete):
                actions = actions.reshape((-1, len(self.action_space.nvec)))
                assert observation.shape[0] == actions.shape[0], \
                    "Error: batch sizes differ for actions and observations."
                # Discrete action probability, over multiple categories
                actions = np.swapaxes(actions, 0, 1)  # swap axis for easier categorical split
                prob = np.prod([proba[np.arange(act.shape[0]), act]
                                         for proba, act in zip(actions_proba, actions)], axis=0)

            elif isinstance(self.action_space, gym.spaces.MultiBinary):
                actions = actions.reshape((-1, self.action_space.n))
                assert observation.shape[0] == actions.shape[0], \
                    "Error: batch sizes differ for actions and observations."
                # Bernoulli action probability, for every action
                prob = np.prod(actions_proba * actions + (1 - actions_proba) * (1 - actions), axis=1)

            elif isinstance(self.action_space, gym.spaces.Box):
                actions = actions.reshape((-1, ) + self.action_space.shape)
                mean, logstd = actions_proba
                std = np.exp(logstd)

                n_elts = np.prod(mean.shape[1:])  # first dimension is batch size
                log_normalizer = n_elts / 2 * np.log(2 * np.pi) + 0.5 * np.sum(logstd, axis=1)

                # Diagonal Gaussian action probability, for every action
                logprob = -np.sum(np.square(actions - mean) / (2 * std), axis=1) - log_normalizer

            else:
                warnings.warn("Warning: action_probability not implemented for {} actions space. Returning None."
                              .format(type(self.action_space).__name__))
                return None

            # Return in space (log or normal) requested by user, converting if necessary
            if logp:
                if logprob is None:
                    logprob = np.log(prob)
                ret = logprob
            else:
                if prob is None:
                    prob = np.exp(logprob)
                ret = prob

            # normalize action proba shape for the different gym spaces
            ret = ret.reshape((-1, 1))
        else:
            ret = actions_proba

        if not vectorized_env:
            if state is not None:
                raise ValueError("Error: The environment must be vectorized when using recurrent policies.")
            ret = ret[0]

        return ret

    def get_parameter_list(self):
        return self.params

    @abstractmethod
    def save(self, save_path, cloudpickle=False):
        pass

    @classmethod
    def load(cls, load_path, env=None, custom_objects=None, **kwargs):
        """
        Load the model from file

        :param load_path: (str or file-like) the saved parameter location
        :param env: (Gym Environment) the new environment to run the loaded model on
            (can be None if you only need prediction from a trained model)
        :param custom_objects: (dict) Dictionary of objects to replace
            upon loading. If a variable is present in this dictionary as a
            key, it will not be deserialized and the corresponding item
            will be used instead. Similar to custom_objects in
            `keras.models.load_model`. Useful when you have an object in
            file that can not be deserialized.
        :param kwargs: extra arguments to change the model when loading
        """
        data, params = cls._load_from_file(load_path, custom_objects=custom_objects)

        if 'policy_kwargs' in kwargs and kwargs['policy_kwargs'] != data['policy_kwargs']:
            raise ValueError("The specified policy kwargs do not equal the stored policy kwargs. "
                             "Stored kwargs: {}, specified kwargs: {}".format(data['policy_kwargs'],
                                                                              kwargs['policy_kwargs']))

        model = cls(policy=data["policy"], env=None, _init_setup_model=False)
        model.__dict__.update(data)
        model.__dict__.update(kwargs)
        model.set_env(env)
        model.setup_model()

        model.load_parameters(params)

        return model


class OffPolicyRLModel(BaseRLModel):
    """
    The base class for off policy RL model

    :param policy: (BasePolicy) Policy object
    :param env: (Gym environment) The environment to learn from
                (if registered in Gym, can be str. Can be None for loading trained models)
    :param replay_buffer: (ReplayBuffer) the type of replay buffer
    :param verbose: (int) the verbosity level: 0 none, 1 training information, 2 tensorflow debug
    :param requires_vec_env: (bool) Does this model require a vectorized environment
    :param policy_base: (BasePolicy) the base policy used by this method
    :param policy_kwargs: (dict) additional arguments to be passed to the policy on creation
    :param seed: (int) Seed for the pseudo-random generators (python, numpy, tensorflow).
        If None (default), use random seed. Note that if you want completely deterministic
        results, you must set `n_cpu_tf_sess` to 1.
    :param n_cpu_tf_sess: (int) The number of threads for TensorFlow operations
        If None, the number of cpu of the current machine will be used.
    """

    def __init__(self, policy, env, replay_buffer=None, _init_setup_model=False, verbose=0, *,
                 requires_vec_env=False, policy_base=None,
                 policy_kwargs=None, seed=None, n_cpu_tf_sess=None):
        super(OffPolicyRLModel, self).__init__(policy, env, verbose=verbose, requires_vec_env=requires_vec_env,
                                               policy_base=policy_base, policy_kwargs=policy_kwargs,
                                               seed=seed, n_cpu_tf_sess=n_cpu_tf_sess)

        self.replay_buffer = replay_buffer

    def is_using_her(self) -> bool:
#        """
#        Check if is using HER
#
#        :return: (bool) Whether is using HER or not
#        """
        # Avoid circular import
        from neorl.rl.baselines.deepq.replay_buffer import HindsightExperienceReplayWrapper
        return isinstance(self.replay_buffer, HindsightExperienceReplayWrapper)

    def replay_buffer_add(self, obs_t, action, reward, obs_tp1, done, info):
#        """
#        Add a new transition to the replay buffer
#
#        :param obs_t: (np.ndarray) the last observation
#        :param action: ([float]) the action
#        :param reward: (float) the reward of the transition
#        :param obs_tp1: (np.ndarray) the new observation
#        :param done: (bool) is the episode done
#        :param info: (dict) extra values used to compute the reward when using HER
#        """
        # Pass info dict when using HER, as it can be used to compute the reward
        kwargs = dict(info=info) if self.is_using_her() else {}
        self.replay_buffer.add(obs_t, action, reward, obs_tp1, float(done), **kwargs)

    @abstractmethod
    def setup_model(self):
        pass

    @abstractmethod
    def learn(self, total_timesteps, callback=None,
              log_interval=100, tb_log_name="run", reset_num_timesteps=True, replay_wrapper=None):
        pass

    @abstractmethod
    def predict(self, observation, state=None, mask=None, deterministic=False):
        pass

    @abstractmethod
    def action_probability(self, observation, state=None, mask=None, actions=None, logp=False):
        pass

    @abstractmethod
    def save(self, save_path, cloudpickle=False):
        pass

    @classmethod
    def load(cls, load_path, env=None, custom_objects=None, **kwargs):
        """
        Load the model from file

        :param load_path: (str or file-like) the saved parameter location
        :param env: (Gym Environment) the new environment to run the loaded model on
            (can be None if you only need prediction from a trained model)
        :param custom_objects: (dict) Dictionary of objects to replace
            upon loading. If a variable is present in this dictionary as a
            key, it will not be deserialized and the corresponding item
            will be used instead. Similar to custom_objects in
            `keras.models.load_model`. Useful when you have an object in
            file that can not be deserialized.
        :param kwargs: extra arguments to change the model when loading
        """
        data, params = cls._load_from_file(load_path, custom_objects=custom_objects)

        if 'policy_kwargs' in kwargs and kwargs['policy_kwargs'] != data['policy_kwargs']:
            raise ValueError("The specified policy kwargs do not equal the stored policy kwargs. "
                             "Stored kwargs: {}, specified kwargs: {}".format(data['policy_kwargs'],
                                                                              kwargs['policy_kwargs']))

        model = cls(policy=data["policy"], env=None, _init_setup_model=False)
        model.__dict__.update(data)
        model.__dict__.update(kwargs)
        model.set_env(env)
        model.setup_model()

        model.load_parameters(params)

        return model


class _UnvecWrapper(VecEnvWrapper):
    def __init__(self, venv):
        """
        Unvectorize a vectorized environment, for vectorized environment that only have one environment

        :param venv: (VecEnv) the vectorized environment to wrap
        """
        super().__init__(venv)
        assert venv.num_envs == 1, "Error: cannot unwrap a environment wrapper that has more than one environment."

    def seed(self, seed=None):
        return self.venv.env_method('seed', seed)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self.venv, attr)

    def __set_attr__(self, attr, value):
        if attr in self.__dict__:
            setattr(self, attr, value)
        else:
            setattr(self.venv, attr, value)

    def compute_reward(self, achieved_goal, desired_goal, _info):
        return float(self.venv.env_method('compute_reward', achieved_goal, desired_goal, _info)[0])

    @staticmethod
    def unvec_obs(obs):
        """
        :param obs: (Union[np.ndarray, dict])
        :return: (Union[np.ndarray, dict])
        """
        if not isinstance(obs, dict):
            return obs[0]
        obs_ = OrderedDict()
        for key in obs.keys():
            obs_[key] = obs[key][0]
        del obs
        return obs_

    def reset(self):
        return self.unvec_obs(self.venv.reset())

    def step_async(self, actions):
        self.venv.step_async([actions])

    def step_wait(self):
        obs, rewards, dones, information = self.venv.step_wait()
        return self.unvec_obs(obs), float(rewards[0]), dones[0], information[0]

    def render(self, mode='human'):
        return self.venv.render(mode=mode)


class SetVerbosity:
    def __init__(self, verbose=0):
        """
        define a region of code for certain level of verbosity

        :param verbose: (int) the verbosity level: 0 none, 1 training information, 2 tensorflow debug
        """
        self.verbose = verbose

    def __enter__(self):
        self.tf_level = os.environ.get('TF_CPP_MIN_LOG_LEVEL', '0')
        self.log_level = logger.get_level()
        self.gym_level = gym.logger.MIN_LEVEL

        if self.verbose <= 1:
            os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

        if self.verbose <= 0:
            logger.set_level(logger.DISABLED)
            gym.logger.set_level(gym.logger.DISABLED)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.verbose <= 1:
            os.environ['TF_CPP_MIN_LOG_LEVEL'] = self.tf_level

        if self.verbose <= 0:
            logger.set_level(self.log_level)
            gym.logger.set_level(self.gym_level)


class TensorboardWriter:
    def __init__(self, graph, tensorboard_log_path, tb_log_name, new_tb_log=True):
        """
        Create a Tensorboard writer for a code segment, and saves it to the log directory as its own run

        :param graph: (Tensorflow Graph) the model graph
        :param tensorboard_log_path: (str) the save path for the log (can be None for no logging)
        :param tb_log_name: (str) the name of the run for tensorboard log
        :param new_tb_log: (bool) whether or not to create a new logging folder for tensorbaord
        """
        self.graph = graph
        self.tensorboard_log_path = tensorboard_log_path
        self.tb_log_name = tb_log_name
        self.writer = None
        self.new_tb_log = new_tb_log

    def __enter__(self):
        if self.tensorboard_log_path is not None:
            latest_run_id = self._get_latest_run_id()
            if self.new_tb_log:
                latest_run_id = latest_run_id + 1
            save_path = os.path.join(self.tensorboard_log_path, "{}_{}".format(self.tb_log_name, latest_run_id))
            self.writer = tf.summary.FileWriter(save_path, graph=self.graph)
        return self.writer

    def _get_latest_run_id(self):
        """
        returns the latest run number for the given log name and log path,
        by finding the greatest number in the directories.

        :return: (int) latest run number
        """
        max_run_id = 0
        for path in glob.glob("{}/{}_[0-9]*".format(self.tensorboard_log_path, self.tb_log_name)):
            file_name = path.split(os.sep)[-1]
            ext = file_name.split("_")[-1]
            if self.tb_log_name == "_".join(file_name.split("_")[:-1]) and ext.isdigit() and int(ext) > max_run_id:
                max_run_id = int(ext)
        return max_run_id

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.writer is not None:
            self.writer.add_graph(self.graph)
            self.writer.flush()



from neorl import PESA

def test_pesa():
    #Define the fitness function
    def FIT(individual):
            """Sphere test objective function.
                    F(x) = sum_{i=1}^d xi^2
                    d=1,2,3,...
                    Range: [-100,100]
                    Minima: 0
            """
            y=sum(x**2 for x in individual)
            return y
    
    #Setup the parameter space (d=5)
    nx=5
    BOUNDS={}
    for i in range(1,nx+1):
            BOUNDS['x'+str(i)]=['float', -100, 100]
    
    npop=60
    pesa=PESA(mode='min', bounds=BOUNDS, fit=FIT, npop=npop, mu=40, alpha_init=0.2, 
              alpha_end=1.0, alpha_backdoor=0.1, ncores=1)
    x0=[[50,50,50,50,50] for i in range(npop)]  #initial guess
    x_best, y_best, pesa_hist=pesa.evolute(ngen=50, x0=x0, verbose=1)

test_pesa()


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Tue Jul 13 21:06:40 2021
#
#@author: Paul Seurin
#"""

########################
#
# Import Packages
#
########################

import gym
from gym.spaces import Discrete, Box, MultiDiscrete
import numpy as np
import bisect
import copy
import matplotlib.pyplot as plt
import random

####################################
#
# Define Traveling Salesman Function
#           (Discrete)
#
####################################

class TSP(gym.Env):

  def __init__ (self,city_loc_list,optimum_tour_city = None, episode_length = None, method = None):                
    self.city_library = {}
    self.city_dist_to_origin = {}
    self.city_clost_to_each_other = {}
    self.city_clost_to_each_other_for_helperfunc = {}
    self.number_of_cities = len(city_loc_list)
    self.state = - 10**6.0 * np.ones((self.number_of_cities,4), dtype = float) # initialize the state space with a single number 
    for i in range(len(city_loc_list)): # initialize the dictionnaries
        self.city_library["%d"%(i + 1)] = [city_loc_list[i][0],city_loc_list[i][1]]
        loc1 = np.array([city_loc_list[i][0],city_loc_list[i][1]])
        dist = np.sqrt(np.sum(np.power(loc1,2)))
        self.city_dist_to_origin["%d"%(i + 1)] = dist # get the distance of each city to the origin and store it in a dictionnary
        close_dist = []
        i_close_dist = []
        for index in range(0,self.number_of_cities):
            if index != i:
                loc2 = np.array([city_loc_list[index][0],city_loc_list[index][1]])
                dist2 = np.sqrt(np.sum(np.power(loc1 - loc2,2)))
                bisect.insort_left(close_dist,dist2) 
                new_index = np.where(close_dist == dist2)[0][0]
                i_close_dist = np.insert(i_close_dist,new_index,"%d"%(index+1))
        self.city_clost_to_each_other["%d"%(i + 1)] = close_dist 
        self.city_clost_to_each_other_for_helperfunc["%d"%(i + 1)] = i_close_dist #  evaluate the closest city to each city and store it in a dictionnary
    
    self.episode_length = episode_length # Number of tsp solved per episodes
    self.city_id = copy.deepcopy(list(self.city_library.keys())) # list with the ids of the cities
    self.city_list = [] # List representing a tour
    
    self.counter = 0 # counter to record a tour
    self.subcounter = 0 # counter to record each tour per episode
    self.method = method #RL method used for optimization
    if self.method not in ['ppo', 'a2c', 'acktr', 'neat', 'dqn', 'acer']:
        raise ValueError("--error: The only methods available are 'ppo', 'a2c', 'acktr', 'neat', 'dqn', or 'acer' not {}".format(self.method))
    if self.method in ['ppo', 'a2c', 'acktr', 'neat']:
        self.action_bounds = [i for i in range(self.number_of_cities,0,-1)]
        self.action_space = MultiDiscrete(self.action_bounds) # action space 
    elif self.method in ['acer', 'dqn']:
        self.action_bounds = self.number_of_cities
        self.action_space = Discrete(self.action_bounds) # action space 
    self.observation_space = Box(low=0, high=self.number_of_cities, shape=(self.number_of_cities * 4,), dtype=int) # state space  
    self.best_tour = -10**6 # variable to store the value of the best tour
    self._iter_episode = 0 # record current episode (for logger)
    self.reset()
    self.done = False
    if optimum_tour_city is not None: # record the (known) best tour map and plot it
        optimum_tour_map = - 10**6.0 * np.ones((self.number_of_cities,4), dtype = float) 
        for increment in range(0,len(list(self.city_library.keys()))):
            optimum_tour_map[:,0][increment] = self.city_library['%d'%(optimum_tour_city[increment])][0]
            optimum_tour_map[:,1][increment] = self.city_library['%d'%(optimum_tour_city[increment])][1] 
        optimum_tour_value = np.ceil(-1 * self.Compute_tour_cost(tour = optimum_tour_map))
        title_map = " Length of Tour : {} \n".format(optimum_tour_value)
        _plot_tour_map(optimum_tour_map,flag = False, name = "Optimum_Tour_map_%d.png"%(len(self.city_id)), title_map = title_map)

  def step(self, x): 
    if self.method in ['ppo', 'a2c', 'acktr', 'neat']:
        self.counter = 0 # initialize the per tour counter
        for action in x: 
            if self.counter == 0:
                self.reset()
            self.counter += 1

            if action <= len(self.city_id) - 1:  
                new_city = self.city_id[action] # recover the city corresponding to action 'action'
            else: # ensure a valid action considering that the action space is fixed
                action = random.randint(0,len(self.city_id) - 1)
                new_city = self.city_id[action]

            if len(np.where(self.state[:,0] != -10**6.0)[0]) != 0:
                non_zero = np.where(self.state[:,0] != -10**6.0)[0]
                max_cost = -10**6.0
                for k in range(len(non_zero)): # place the city chosen in the location that reduces the partial tour
                    temp_state_0 = copy.deepcopy(self.state[:,0])
                    temp_state_1 = copy.deepcopy(self.state[:,1])
                    self.state[:,0] = np.delete(np.insert(self.state[:,0],k,self.city_library[new_city][0]),-1)
                    self.state[:,1] = np.delete(np.insert(self.state[:,1],k,self.city_library[new_city][1]),-1)
                    temp_score = self.Compute_tour_cost()# evalaute the cost of a partial tour 
                    if temp_score > max_cost:
                        best_k = k
                        max_cost = temp_score
                    self.state[:,0] = temp_state_0
                    self.state[:,1] = temp_state_1

                # fill the state space
                self.state[:,0] = np.delete(np.insert(self.state[:,0],best_k,self.city_library[new_city][0]),-1) 
                self.state[:,1] = np.delete(np.insert(self.state[:,1],best_k,self.city_library[new_city][1]),-1)
                self.state[:,2][self.counter - 1] = self.city_clost_to_each_other[new_city][0]
                self.state[:,3][self.counter - 1] = self.city_dist_to_origin[new_city]

            else: # place the first city and fill the state space
                self.state[:,0][self.counter - 1] = self.city_library[new_city][0]
                self.state[:,1][self.counter - 1] = self.city_library[new_city][1] 
                self.state[:,2][self.counter - 1] = self.city_clost_to_each_other[new_city][0]
                self.state[:,3][self.counter - 1] = self.city_dist_to_origin[new_city]

            index1 = self.city_id.index(self.city_id[action]) # remove the city chosen by action : 'action' from the action space
            self.city_id.pop(index1)
            self.city_list.append(new_city) # sequence of cities in the tour (for neorl callback)

        reward = self._get_stats()
        self.subcounter += 1 
        if self.subcounter != self.episode_length + 1: # generate the new tour
            self.city_id = copy.deepcopy(list(self.city_library.keys()))
            individual = self.city_list
            self.city_list = []
            self.state[:,0] = - 10**6.0 * np.ones(self.number_of_cities)
            self.state[:,1] = - 10**6.0 * np.ones(self.number_of_cities)
            self.state[:,2] = - 10**6.0 * np.ones(self.number_of_cities)
            self.state[:,3] = - 10**6.0 * np.ones(self.number_of_cities)      
    
    elif self.method in ['acer', 'dqn']:
        action = x
        if action <= len(self.city_id) - 1:  
            new_city = self.city_id[action] # recover the city corresponding to action 'action'
        else: # ensure a valid action considering that the action space is fixed
            action = random.randint(0,len(self.city_id) - 1)
            new_city = self.city_id[action]
        self.counter += 1 # --- increment the global counter
        if len(np.where(self.state[:,0] != -10**6.0)[0]) != 0:
            non_zero = np.where(self.state[:,0] != -10**6.0)[0]
            max_cost = -10**6.0
            for k in range(len(non_zero)): # place the city chosen in the location that reduces the partial tour
                temp_state_0 = copy.deepcopy(self.state[:,0])
                temp_state_1 = copy.deepcopy(self.state[:,1])
                self.state[:,0] = np.delete(np.insert(self.state[:,0],k,self.city_library[new_city][0]),-1)
                self.state[:,1] = np.delete(np.insert(self.state[:,1],k,self.city_library[new_city][1]),-1)
                temp_score = self.Compute_tour_cost()# evalaute the cost of a partial tour 
                if temp_score > max_cost:
                    best_k = k
                    max_cost = temp_score
                self.state[:,0] = temp_state_0
                self.state[:,1] = temp_state_1
            # fill the state space
            self.state[:,0] = np.delete(np.insert(self.state[:,0],best_k,self.city_library[new_city][0]),-1) 
            self.state[:,1] = np.delete(np.insert(self.state[:,1],best_k,self.city_library[new_city][1]),-1)
            self.state[:,2][self.counter - 1] = self.city_clost_to_each_other[new_city][0]
            self.state[:,3][self.counter - 1] = self.city_dist_to_origin[new_city]
        else:
            self.state[:,0][self.counter - 1] = self.city_library[new_city][0]
            self.state[:,1][self.counter - 1] = self.city_library[new_city][1] 
            self.state[:,2][self.counter - 1] = self.city_clost_to_each_other[new_city][0]
            self.state[:,3][self.counter - 1] = self.city_dist_to_origin[new_city]


        index1 = self.city_id.index(self.city_id[action]) # remove the city chosen by action : 'action' from the action space
        self.city_id.pop(index1)
        self.city_list.append(new_city) # sequence of cities in the tour (for neorl callback)
        reward = self._get_stats(flag = False#self.state = - 10**6.0 * np.ones((self.number_of_cities,4), dtype = float)
    ) * -10**3 # evaluate a partial tour. Increase the reward for it to be always bigger than a full tour

        if len(np.where(self.state[:,0] == -10**6.0)[0]) == 0:# a tour is complete.
            reward = self._get_stats()
            self.subcounter += 1
            self.counter = 0
            if self.subcounter != self.episode_length + 1: # generate the new tour 
                self.city_id = copy.deepcopy(list(self.city_library.keys()))
                individual = self.city_list
                self.city_list = []
                #self.state = - 10**6.0* * np.ones((self.number_of_cities,4), dtype = float)# reinitialize the state space
                self.state[:,0] = - 10**6.0 * np.ones(self.number_of_cities)
                self.state[:,1] = - 10**6.0 * np.ones(self.number_of_cities)
                self.state[:,2] = - 10**6.0 * np.ones(self.number_of_cities)
                self.state[:,3] = - 10**6.0 * np.ones(self.number_of_cities)
        else:
            individual = self.city_list # a tour is not complete. No individuals are return (for neorl callback function)

    if self.subcounter == self.episode_length:# episode terminates
        self.done = True
        self._iter_episode += 1
        self.subcounter = 0
    return ([self.state.flatten(),reward, self.done, {'x':individual}])

  def reset(self):
    self.done = False
    self.city_id = copy.deepcopy(list(self.city_library.keys()))
    self.state[:,0] = - 10**6.0 * np.ones(self.number_of_cities)
    self.state[:,1] = - 10**6.0 * np.ones(self.number_of_cities)
    self.state[:,2] = - 10**6.0 * np.ones(self.number_of_cities)
    self.state[:,3] = - 10**6.0 * np.ones(self.number_of_cities)
    return (self.state.flatten())

  def Compute_tour_cost(self, tour = None):
    if tour is None:
        tour = copy.deepcopy(self.state)
    cost = 0
    limit_tour = np.where(tour[:,0] == -10**6.0)[0] # a tour ends with the traveling salesman circling back to the initial city
    if len(limit_tour) > 0:
        limit_tour = limit_tour[0] - 1
    else:
        limit_tour = self.number_of_cities - 1  
    for increment in range(0,limit_tour): # compute euclidean distance from cities to cities
        loc1 = np.array([tour[:,0][increment],tour[:,1][increment]])
        loc2 = np.array([tour[:,0][increment + 1],tour[:,1][increment + 1]])
        dist = np.sqrt(np.sum(np.power(loc1 - loc2,2)))
        cost += int(round(dist))
    loc1 = np.array([tour[:,0][limit_tour],tour[:,1][limit_tour]])
    loc2 = np.array([tour[:,0][0],tour[:,1][0]])
    dist = np.sqrt(np.sum(np.power(loc1 - loc2,2)))
    cost += int(round(dist)) 
    score  =  - cost 
    return score    

  def _get_stats(self,flag = True):
      score = self.Compute_tour_cost() 
      if flag: # compute the best tour and plot it only for complete tour
        if score - self.best_tour > 0:
            self.best_tour = score
            current_tour_cost = -1 * self.Compute_tour_cost(tour = self.state)
            title_map = " Episode: {}, Length of Tour : {} \n".format(self._iter_episode + 1, np.ceil(current_tour_cost))
            _plot_tour_map(self.state,flag = False, name = "Best_Tour_map_%d_%s.png"%(self.number_of_cities,self.method), title_map = title_map)

      return score
            
  def render(self, mode = 'human'):
    pass

def _plot_tour_map(tour_embed,dirname_2 = ".", flag = False, name ="Tour_map.png", title_map = None):
    """
      Plot a Tour stored in : tour_embed
    """
    Full_map = copy.deepcopy(tour_embed)
    X = np.append(Full_map[:,0], Full_map[:,0][0])
    Y = np.append(Full_map[:,1], Full_map[:,1][0])
    X_or = Full_map[:,0]
    Y_or = Full_map[:,1]
    plt.figure()
    if title_map is not None:
        plt.title(title_map)
    else:
        plt.title('Tour map')
    plt.ylabel('y [m]')
    plt.xlabel('x [m]')
    plt.scatter(X_or[0],Y_or[0], color = 'black',label = "Departure")
    plt.scatter(X_or[1:],Y_or[1:], color = 'orange' ,label = "Cities")
    plt.plot(X,Y, "-", label = "Tour")
    plt.legend()
    plt.savefig("%s/%s"%(dirname_2,name),format='png' ,dpi=300, bbox_inches="tight")
    plt.close()



import warnings
from itertools import zip_longest
from abc import ABC, abstractmethod

import numpy as np
import tensorflow as tf
from gym.spaces import Discrete

from neorl.rl.baselines.shared.tf_util import batch_to_seq, seq_to_batch
from neorl.rl.baselines.shared.tf_layers import conv, linear, conv_to_fc, lstm
from neorl.rl.baselines.shared.distributions import make_proba_dist_type, CategoricalProbabilityDistribution, \
    MultiCategoricalProbabilityDistribution, DiagGaussianProbabilityDistribution, BernoulliProbabilityDistribution
from neorl.rl.baselines.shared.input import observation_input


def nature_cnn(scaled_images, **kwargs):
    """
    CNN from Nature paper.

    :param scaled_images: (TensorFlow Tensor) Image input placeholder
    :param kwargs: (dict) Extra keywords parameters for the convolutional layers of the CNN
    :return: (TensorFlow Tensor) The CNN output layer
    """
    activ = tf.nn.relu
    layer_1 = activ(conv(scaled_images, 'c1', n_filters=32, filter_size=8, stride=4, init_scale=np.sqrt(2), **kwargs))
    layer_2 = activ(conv(layer_1, 'c2', n_filters=64, filter_size=4, stride=2, init_scale=np.sqrt(2), **kwargs))
    layer_3 = activ(conv(layer_2, 'c3', n_filters=64, filter_size=3, stride=1, init_scale=np.sqrt(2), **kwargs))
    layer_3 = conv_to_fc(layer_3)
    return activ(linear(layer_3, 'fc1', n_hidden=512, init_scale=np.sqrt(2)))


def mlp_extractor(flat_observations, net_arch, act_fun):
    """
    Constructs an MLP that receives observations as an input and outputs a latent representation for the policy and
    a value network. The ``net_arch`` parameter allows to specify the amount and size of the hidden layers and how many
    of them are shared between the policy network and the value network. It is assumed to be a list with the following
    structure:

    1. An arbitrary length (zero allowed) number of integers each specifying the number of units in a shared layer.
       If the number of ints is zero, there will be no shared layers.
    2. An optional dict, to specify the following non-shared layers for the value network and the policy network.
       It is formatted like ``dict(vf=[<value layer sizes>], pi=[<policy layer sizes>])``.
       If it is missing any of the keys (pi or vf), no non-shared layers (empty list) is assumed.

    For example to construct a network with one shared layer of size 55 followed by two non-shared layers for the value
    network of size 255 and a single non-shared layer of size 128 for the policy network, the following layers_spec
    would be used: ``[55, dict(vf=[255, 255], pi=[128])]``. A simple shared network topology with two layers of size 128
    would be specified as [128, 128].

    :param flat_observations: (tf.Tensor) The observations to base policy and value function on.
    :param net_arch: ([int or dict]) The specification of the policy and value networks.
        See above for details on its formatting.
    :param act_fun: (tf function) The activation function to use for the networks.
    :return: (tf.Tensor, tf.Tensor) latent_policy, latent_value of the specified network.
        If all layers are shared, then ``latent_policy == latent_value``
    """
    latent = flat_observations
    policy_only_layers = []  # Layer sizes of the network that only belongs to the policy network
    value_only_layers = []  # Layer sizes of the network that only belongs to the value network

    # Iterate through the shared layers and build the shared parts of the network
    for idx, layer in enumerate(net_arch):
        # CHANGE: added following print statements
        print('Adding SHARED layer number --', idx, '-- with ', layer, 'nodes')
        if isinstance(layer, int):  # Check that this is a shared layer
            layer_size = layer
            latent = act_fun(linear(latent, "shared_fc{}".format(idx), layer_size, init_scale=np.sqrt(2)))
        else:
            assert isinstance(layer, dict), "Error: the net_arch list can only contain ints and dicts"
            if 'pi' in layer:
                assert isinstance(layer['pi'], list), "Error: net_arch[-1]['pi'] must contain a list of integers."
                policy_only_layers = layer['pi']

            if 'vf' in layer:
                assert isinstance(layer['vf'], list), "Error: net_arch[-1]['vf'] must contain a list of integers."
                value_only_layers = layer['vf']
            break  # From here on the network splits up in policy and value network

    # Build the non-shared part of the network
    latent_policy = latent
    latent_value = latent
    for idx, (pi_layer_size, vf_layer_size) in enumerate(zip_longest(policy_only_layers, value_only_layers)):
        # CHANGE: added following print statements
        print('Adding NON-SHARED layer number --', idx, '-- with ', layer, 'nodes')
        if pi_layer_size is not None:
            assert isinstance(pi_layer_size, int), "Error: net_arch[-1]['pi'] must only contain integers."
            latent_policy = act_fun(linear(latent_policy, "pi_fc{}".format(idx), pi_layer_size, init_scale=np.sqrt(2)))

        if vf_layer_size is not None:
            assert isinstance(vf_layer_size, int), "Error: net_arch[-1]['vf'] must only contain integers."
            latent_value = act_fun(linear(latent_value, "vf_fc{}".format(idx), vf_layer_size, init_scale=np.sqrt(2)))

    return latent_policy, latent_value


class BasePolicy(ABC):
    """
    The base policy object

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batches to run (n_envs * n_steps)
    :param reuse: (bool) If the policy is reusable or not
    :param scale: (bool) whether or not to scale the input
    :param obs_phs: (TensorFlow Tensor, TensorFlow Tensor) a tuple containing an override for observation placeholder
        and the processed observation placeholder respectively
    :param add_action_ph: (bool) whether or not to create an action placeholder
    """

    recurrent = False

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=False, scale=False,
                 obs_phs=None, add_action_ph=False):
        self.n_env = n_env
        self.n_steps = n_steps
        self.n_batch = n_batch
        with tf.variable_scope("input", reuse=False):
            if obs_phs is None:
                self._obs_ph, self._processed_obs = observation_input(ob_space, n_batch, scale=scale)
            else:
                self._obs_ph, self._processed_obs = obs_phs

            self._action_ph = None
            if add_action_ph:
                self._action_ph = tf.placeholder(dtype=ac_space.dtype, shape=(n_batch,) + ac_space.shape,
                                                 name="action_ph")
        self.sess = sess
        self.reuse = reuse
        self.ob_space = ob_space
        self.ac_space = ac_space

    @property
    def is_discrete(self):
        """bool: is action space discrete."""
        return isinstance(self.ac_space, Discrete)

    @property
    def initial_state(self):
        """
        The initial state of the policy. For feedforward policies, None. For a recurrent policy,
        a NumPy array of shape (self.n_env, ) + state_shape.
        """
        assert not self.recurrent, "When using recurrent policies, you must overwrite `initial_state()` method"
        return None

    @property
    def obs_ph(self):
        """tf.Tensor: placeholder for observations, shape (self.n_batch, ) + self.ob_space.shape."""
        return self._obs_ph

    @property
    def processed_obs(self):
        """tf.Tensor: processed observations, shape (self.n_batch, ) + self.ob_space.shape.

        The form of processing depends on the type of the observation space, and the parameters
        whether scale is passed to the constructor; see observation_input for more information."""
        return self._processed_obs

    @property
    def action_ph(self):
        """tf.Tensor: placeholder for actions, shape (self.n_batch, ) + self.ac_space.shape."""
        return self._action_ph

    @staticmethod
    def _kwargs_check(feature_extraction, kwargs):
        """
        Ensure that the user is not passing wrong keywords
        when using policy_kwargs.

        :param feature_extraction: (str)
        :param kwargs: (dict)
        """
        # When using policy_kwargs parameter on model creation,
        # all keywords arguments must be consumed by the policy constructor except
        # the ones for the cnn_extractor network (cf nature_cnn()), where the keywords arguments
        # are not passed explicitly (using **kwargs to forward the arguments)
        # that's why there should be not kwargs left when using the mlp_extractor
        # (in that case the keywords arguments are passed explicitly)
        if feature_extraction == 'mlp' and len(kwargs) > 0:
            raise ValueError("Unknown keywords for policy: {}".format(kwargs))

    @abstractmethod
    def step(self, obs, state=None, mask=None):
        """
        Returns the policy for a single step

        :param obs: ([float] or [int]) The current observation of the environment
        :param state: ([float]) The last states (used in recurrent policies)
        :param mask: ([float]) The last masks (used in recurrent policies)
        :return: ([float], [float], [float], [float]) actions, values, states, neglogp
        """
        raise NotImplementedError

    @abstractmethod
    def proba_step(self, obs, state=None, mask=None):
        """
        Returns the action probability for a single step

        :param obs: ([float] or [int]) The current observation of the environment
        :param state: ([float]) The last states (used in recurrent policies)
        :param mask: ([float]) The last masks (used in recurrent policies)
        :return: ([float]) the action probability
        """
        raise NotImplementedError


class ActorCriticPolicy(BasePolicy):
    """
    Policy object that implements actor critic

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param reuse: (bool) If the policy is reusable or not
    :param scale: (bool) whether or not to scale the input
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=False, scale=False):
        super(ActorCriticPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=reuse,
                                                scale=scale)
        self._pdtype = make_proba_dist_type(ac_space)
        self._policy = None
        self._proba_distribution = None
        self._value_fn = None
        self._action = None
        self._deterministic_action = None

    def _setup_init(self):
        """Sets up the distributions, actions, and value."""
        with tf.variable_scope("output", reuse=True):
            assert self.policy is not None and self.proba_distribution is not None and self.value_fn is not None
            self._action = self.proba_distribution.sample()
            self._deterministic_action = self.proba_distribution.mode()
            self._neglogp = self.proba_distribution.neglogp(self.action)
            if isinstance(self.proba_distribution, CategoricalProbabilityDistribution):
                self._policy_proba = tf.nn.softmax(self.policy)
            elif isinstance(self.proba_distribution, DiagGaussianProbabilityDistribution):
                self._policy_proba = [self.proba_distribution.mean, self.proba_distribution.std]
            elif isinstance(self.proba_distribution, BernoulliProbabilityDistribution):
                self._policy_proba = tf.nn.sigmoid(self.policy)
            elif isinstance(self.proba_distribution, MultiCategoricalProbabilityDistribution):
                self._policy_proba = [tf.nn.softmax(categorical.flatparam())
                                     for categorical in self.proba_distribution.categoricals]
            else:
                self._policy_proba = []  # it will return nothing, as it is not implemented
            self._value_flat = self.value_fn[:, 0]

    @property
    def pdtype(self):
        """ProbabilityDistributionType: type of the distribution for stochastic actions."""
        return self._pdtype

    @property
    def policy(self):
        """tf.Tensor: policy output, e.g. logits."""
        return self._policy

    @property
    def proba_distribution(self):
        """ProbabilityDistribution: distribution of stochastic actions."""
        return self._proba_distribution

    @property
    def value_fn(self):
        """tf.Tensor: value estimate, of shape (self.n_batch, 1)"""
        return self._value_fn

    @property
    def value_flat(self):
        """tf.Tensor: value estimate, of shape (self.n_batch, )"""
        return self._value_flat

    @property
    def action(self):
        """tf.Tensor: stochastic action, of shape (self.n_batch, ) + self.ac_space.shape."""
        return self._action

    @property
    def deterministic_action(self):
        """tf.Tensor: deterministic action, of shape (self.n_batch, ) + self.ac_space.shape."""
        return self._deterministic_action

    @property
    def neglogp(self):
        """tf.Tensor: negative log likelihood of the action sampled by self.action."""
        return self._neglogp

    @property
    def policy_proba(self):
        """tf.Tensor: parameters of the probability distribution. Depends on pdtype."""
        return self._policy_proba

    @abstractmethod
    def step(self, obs, state=None, mask=None, deterministic=False):
        """
        Returns the policy for a single step

        :param obs: ([float] or [int]) The current observation of the environment
        :param state: ([float]) The last states (used in recurrent policies)
        :param mask: ([float]) The last masks (used in recurrent policies)
        :param deterministic: (bool) Whether or not to return deterministic actions.
        :return: ([float], [float], [float], [float]) actions, values, states, neglogp
        """
        raise NotImplementedError

    @abstractmethod
    def value(self, obs, state=None, mask=None):
        """
        Returns the value for a single step

        :param obs: ([float] or [int]) The current observation of the environment
        :param state: ([float]) The last states (used in recurrent policies)
        :param mask: ([float]) The last masks (used in recurrent policies)
        :return: ([float]) The associated value of the action
        """
        raise NotImplementedError


class RecurrentActorCriticPolicy(ActorCriticPolicy):
    """
    Actor critic policy object uses a previous state in the computation for the current step.
    NOTE: this class is not limited to recurrent neural network policies,
    see https://github.com/hill-a/stable-baselines/issues/241

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param state_shape: (tuple<int>) shape of the per-environment state space.
    :param reuse: (bool) If the policy is reusable or not
    :param scale: (bool) whether or not to scale the input
    """

    recurrent = True

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch,
                 state_shape, reuse=False, scale=False):
        super(RecurrentActorCriticPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps,
                                                         n_batch, reuse=reuse, scale=scale)

        with tf.variable_scope("input", reuse=False):
            self._dones_ph = tf.placeholder(tf.float32, (n_batch, ), name="dones_ph")  # (done t-1)
            state_ph_shape = (self.n_env, ) + tuple(state_shape)
            self._states_ph = tf.placeholder(tf.float32, state_ph_shape, name="states_ph")

        initial_state_shape = (self.n_env, ) + tuple(state_shape)
        self._initial_state = np.zeros(initial_state_shape, dtype=np.float32)

    @property
    def initial_state(self):
        return self._initial_state

    @property
    def dones_ph(self):
        """tf.Tensor: placeholder for whether episode has terminated (done), shape (self.n_batch, ).
        Internally used to reset the state before the next episode starts."""
        return self._dones_ph

    @property
    def states_ph(self):
        """tf.Tensor: placeholder for states, shape (self.n_env, ) + state_shape."""
        return self._states_ph

    @abstractmethod
    def value(self, obs, state=None, mask=None):
        """
        Cf base class doc.
        """
        raise NotImplementedError


class LstmPolicy(RecurrentActorCriticPolicy):
    """
    Policy object that implements actor critic, using LSTMs.

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param n_lstm: (int) The number of LSTM cells (for recurrent policies)
    :param reuse: (bool) If the policy is reusable or not
    :param layers: ([int]) The size of the Neural network before the LSTM layer  (if None, default to [64, 64])
    :param net_arch: (list) Specification of the actor-critic policy network architecture. Notation similar to the
        format described in mlp_extractor but with additional support for a 'lstm' entry in the shared network part.
    :param act_fun: (tf.func) the activation function to use in the neural network.
    :param cnn_extractor: (function (TensorFlow Tensor, ``**kwargs``): (TensorFlow Tensor)) the CNN feature extraction
    :param layer_norm: (bool) Whether or not to use layer normalizing LSTMs
    :param feature_extraction: (str) The feature extraction type ("cnn" or "mlp")
    :param kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    recurrent = True

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm=256, reuse=False, layers=None,
                 net_arch=None, act_fun=tf.tanh, cnn_extractor=nature_cnn, layer_norm=False, feature_extraction="cnn",
                 **kwargs):
        # state_shape = [n_lstm * 2] dim because of the cell and hidden states of the LSTM
        super(LstmPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch,
                                         state_shape=(2 * n_lstm, ), reuse=reuse,
                                         scale=(feature_extraction == "cnn"))

        self._kwargs_check(feature_extraction, kwargs)

        if net_arch is None:  # Legacy mode
            if layers is None:
                layers = [64, 64]
            else:
                warnings.warn("The layers parameter is deprecated. Use the net_arch parameter instead.")

            with tf.variable_scope("model", reuse=reuse):
                if feature_extraction == "cnn":
                    extracted_features = cnn_extractor(self.processed_obs, **kwargs)
                else:
                    extracted_features = tf.layers.flatten(self.processed_obs)
                    for i, layer_size in enumerate(layers):
                        extracted_features = act_fun(linear(extracted_features, 'pi_fc' + str(i), n_hidden=layer_size,
                                                            init_scale=np.sqrt(2)))
                input_sequence = batch_to_seq(extracted_features, self.n_env, n_steps)
                masks = batch_to_seq(self.dones_ph, self.n_env, n_steps)
                rnn_output, self.snew = lstm(input_sequence, masks, self.states_ph, 'lstm1', n_hidden=n_lstm,
                                             layer_norm=layer_norm)
                rnn_output = seq_to_batch(rnn_output)
                value_fn = linear(rnn_output, 'vf', 1)

                self._proba_distribution, self._policy, self.q_value = \
                    self.pdtype.proba_distribution_from_latent(rnn_output, rnn_output)

            self._value_fn = value_fn
        else:  # Use the new net_arch parameter
            if layers is not None:
                warnings.warn("The new net_arch parameter overrides the deprecated layers parameter.")
            if feature_extraction == "cnn":
                raise NotImplementedError()

            with tf.variable_scope("model", reuse=reuse):
                latent = tf.layers.flatten(self.processed_obs)
                policy_only_layers = []  # Layer sizes of the network that only belongs to the policy network
                value_only_layers = []  # Layer sizes of the network that only belongs to the value network

                # Iterate through the shared layers and build the shared parts of the network
                lstm_layer_constructed = False
                for idx, layer in enumerate(net_arch):
                    if isinstance(layer, int):  # Check that this is a shared layer
                        layer_size = layer
                        latent = act_fun(linear(latent, "shared_fc{}".format(idx), layer_size, init_scale=np.sqrt(2)))
                    elif layer == "lstm":
                        if lstm_layer_constructed:
                            raise ValueError("The net_arch parameter must only contain one occurrence of 'lstm'!")
                        input_sequence = batch_to_seq(latent, self.n_env, n_steps)
                        masks = batch_to_seq(self.dones_ph, self.n_env, n_steps)
                        rnn_output, self.snew = lstm(input_sequence, masks, self.states_ph, 'lstm1', n_hidden=n_lstm,
                                                     layer_norm=layer_norm)
                        latent = seq_to_batch(rnn_output)
                        lstm_layer_constructed = True
                    else:
                        assert isinstance(layer, dict), "Error: the net_arch list can only contain ints and dicts"
                        if 'pi' in layer:
                            assert isinstance(layer['pi'],
                                              list), "Error: net_arch[-1]['pi'] must contain a list of integers."
                            policy_only_layers = layer['pi']

                        if 'vf' in layer:
                            assert isinstance(layer['vf'],
                                              list), "Error: net_arch[-1]['vf'] must contain a list of integers."
                            value_only_layers = layer['vf']
                        break  # From here on the network splits up in policy and value network

                # Build the non-shared part of the policy-network
                latent_policy = latent
                for idx, pi_layer_size in enumerate(policy_only_layers):
                    if pi_layer_size == "lstm":
                        raise NotImplementedError("LSTMs are only supported in the shared part of the policy network.")
                    assert isinstance(pi_layer_size, int), "Error: net_arch[-1]['pi'] must only contain integers."
                    latent_policy = act_fun(
                        linear(latent_policy, "pi_fc{}".format(idx), pi_layer_size, init_scale=np.sqrt(2)))

                # Build the non-shared part of the value-network
                latent_value = latent
                for idx, vf_layer_size in enumerate(value_only_layers):
                    if vf_layer_size == "lstm":
                        raise NotImplementedError("LSTMs are only supported in the shared part of the value function "
                                                  "network.")
                    assert isinstance(vf_layer_size, int), "Error: net_arch[-1]['vf'] must only contain integers."
                    latent_value = act_fun(
                        linear(latent_value, "vf_fc{}".format(idx), vf_layer_size, init_scale=np.sqrt(2)))

                if not lstm_layer_constructed:
                    raise ValueError("The net_arch parameter must contain at least one occurrence of 'lstm'!")

                self._value_fn = linear(latent_value, 'vf', 1)
                # TODO: why not init_scale = 0.001 here like in the feedforward
                self._proba_distribution, self._policy, self.q_value = \
                    self.pdtype.proba_distribution_from_latent(latent_policy, latent_value)
        self._setup_init()

    def step(self, obs, state=None, mask=None, deterministic=False):
        if deterministic:
            return self.sess.run([self.deterministic_action, self.value_flat, self.snew, self.neglogp],
                                 {self.obs_ph: obs, self.states_ph: state, self.dones_ph: mask})
        else:
            return self.sess.run([self.action, self.value_flat, self.snew, self.neglogp],
                                 {self.obs_ph: obs, self.states_ph: state, self.dones_ph: mask})

    def proba_step(self, obs, state=None, mask=None):
        return self.sess.run(self.policy_proba, {self.obs_ph: obs, self.states_ph: state, self.dones_ph: mask})

    def value(self, obs, state=None, mask=None):
        return self.sess.run(self.value_flat, {self.obs_ph: obs, self.states_ph: state, self.dones_ph: mask})


class FeedForwardPolicy(ActorCriticPolicy):
    """
    Policy object that implements actor critic, using a feed forward neural network.

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param reuse: (bool) If the policy is reusable or not
    :param layers: ([int]) (deprecated, use net_arch instead) The size of the Neural network for the policy
        (if None, default to [64, 64])
    :param net_arch: (list) Specification of the actor-critic policy network architecture (see mlp_extractor
        documentation for details).
    :param act_fun: (tf.func) the activation function to use in the neural network.
    :param cnn_extractor: (function (TensorFlow Tensor, ``**kwargs``): (TensorFlow Tensor)) the CNN feature extraction
    :param feature_extraction: (str) The feature extraction type ("cnn" or "mlp")
    :param kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=False, layers=None, net_arch=None,
                 act_fun=tf.tanh, cnn_extractor=nature_cnn, feature_extraction="cnn", **kwargs):
        super(FeedForwardPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=reuse,
                                                scale=(feature_extraction == "cnn"))

        self._kwargs_check(feature_extraction, kwargs)

        if layers is not None:
            warnings.warn("Usage of the `layers` parameter is deprecated! Use net_arch instead "
                          "(it has a different semantics though).", DeprecationWarning)
            if net_arch is not None:
                warnings.warn("The new `net_arch` parameter overrides the deprecated `layers` parameter!",
                              DeprecationWarning)

        if net_arch is None:
            if layers is None:
                layers = [64, 64]
            net_arch = [dict(vf=layers, pi=layers)]

        with tf.variable_scope("model", reuse=reuse):
            if feature_extraction == "cnn":
                pi_latent = vf_latent = cnn_extractor(self.processed_obs, **kwargs)
            else:
                pi_latent, vf_latent = mlp_extractor(tf.layers.flatten(self.processed_obs), net_arch, act_fun)

            self._value_fn = linear(vf_latent, 'vf', 1)

            self._proba_distribution, self._policy, self.q_value = \
                self.pdtype.proba_distribution_from_latent(pi_latent, vf_latent, init_scale=0.01)

        self._setup_init()

    def step(self, obs, state=None, mask=None, deterministic=False):
        if deterministic:
            action, value, neglogp = self.sess.run([self.deterministic_action, self.value_flat, self.neglogp],
                                                   {self.obs_ph: obs})
        else:
            action, value, neglogp = self.sess.run([self.action, self.value_flat, self.neglogp],
                                                   {self.obs_ph: obs})
        return action, value, self.initial_state, neglogp

    def proba_step(self, obs, state=None, mask=None):
        return self.sess.run(self.policy_proba, {self.obs_ph: obs})

    def value(self, obs, state=None, mask=None):
        return self.sess.run(self.value_flat, {self.obs_ph: obs})


class CnnPolicy(FeedForwardPolicy):
    """
    Policy object that implements actor critic, using a CNN (the nature CNN)

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param reuse: (bool) If the policy is reusable or not
    :param _kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=False, **_kwargs):
        super(CnnPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse,
                                        feature_extraction="cnn", **_kwargs)


class CnnLstmPolicy(LstmPolicy):
    """
    Policy object that implements actor critic, using LSTMs with a CNN feature extraction

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param n_lstm: (int) The number of LSTM cells (for recurrent policies)
    :param reuse: (bool) If the policy is reusable or not
    :param kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm=256, reuse=False, **_kwargs):
        super(CnnLstmPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm, reuse,
                                            layer_norm=False, feature_extraction="cnn", **_kwargs)


class CnnLnLstmPolicy(LstmPolicy):
    """
    Policy object that implements actor critic, using a layer normalized LSTMs with a CNN feature extraction

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param n_lstm: (int) The number of LSTM cells (for recurrent policies)
    :param reuse: (bool) If the policy is reusable or not
    :param kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm=256, reuse=False, **_kwargs):
        super(CnnLnLstmPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm, reuse,
                                              layer_norm=True, feature_extraction="cnn", **_kwargs)


class MlpPolicy(FeedForwardPolicy):
    """
    Policy object that implements actor critic, using a MLP (2 layers of 64)

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param reuse: (bool) If the policy is reusable or not
    :param _kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse=False, **_kwargs):
        super(MlpPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, reuse,
                                        feature_extraction="mlp", **_kwargs)


class MlpLstmPolicy(LstmPolicy):
    """
    Policy object that implements actor critic, using LSTMs with a MLP feature extraction

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param n_lstm: (int) The number of LSTM cells (for recurrent policies)
    :param reuse: (bool) If the policy is reusable or not
    :param kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm=256, reuse=False, **_kwargs):
        super(MlpLstmPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm, reuse,
                                            layer_norm=False, feature_extraction="mlp", **_kwargs)


class MlpLnLstmPolicy(LstmPolicy):
    """
    Policy object that implements actor critic, using a layer normalized LSTMs with a MLP feature extraction

    :param sess: (TensorFlow session) The current TensorFlow session
    :param ob_space: (Gym Space) The observation space of the environment
    :param ac_space: (Gym Space) The action space of the environment
    :param n_env: (int) The number of environments to run
    :param n_steps: (int) The number of steps to run for each environment
    :param n_batch: (int) The number of batch to run (n_envs * n_steps)
    :param n_lstm: (int) The number of LSTM cells (for recurrent policies)
    :param reuse: (bool) If the policy is reusable or not
    :param kwargs: (dict) Extra keyword arguments for the nature CNN feature extraction
    """

    def __init__(self, sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm=256, reuse=False, **_kwargs):
        super(MlpLnLstmPolicy, self).__init__(sess, ob_space, ac_space, n_env, n_steps, n_batch, n_lstm, reuse,
                                              layer_norm=True, feature_extraction="mlp", **_kwargs)


_policy_registry = {
    ActorCriticPolicy: {
        "CnnPolicy": CnnPolicy,
        "CnnLstmPolicy": CnnLstmPolicy,
        "CnnLnLstmPolicy": CnnLnLstmPolicy,
        "MlpPolicy": MlpPolicy,
        "MlpLstmPolicy": MlpLstmPolicy,
        "MlpLnLstmPolicy": MlpLnLstmPolicy,
    }
}


def get_policy_from_name(base_policy_type, name):
    """
    returns the registed policy from the base type and name

    :param base_policy_type: (BasePolicy) the base policy object
    :param name: (str) the policy name
    :return: (base_policy_type) the policy
    """
    if base_policy_type not in _policy_registry:
        raise ValueError("Error: the policy type {} is not registered!".format(base_policy_type))
    if name not in _policy_registry[base_policy_type]:
        raise ValueError("Error: unknown policy type {}, the only registed policy type are: {}!"
                         .format(name, list(_policy_registry[base_policy_type].keys())))
    return _policy_registry[base_policy_type][name]


def register_policy(name, policy):
    """
    returns the registed policy from the base type and name

    :param name: (str) the policy name
    :param policy: (subclass of BasePolicy) the policy
    """
    sub_class = None
    for cls in BasePolicy.__subclasses__():
        if issubclass(policy, cls):
            sub_class = cls
            break
    if sub_class is None:
        raise ValueError("Error: the policy {} is not of any known subclasses of BasePolicy!".format(policy))

    if sub_class not in _policy_registry:
        _policy_registry[sub_class] = {}
    if name in _policy_registry[sub_class]:
        raise ValueError("Error: the name {} is alreay registered for a different policy, will not override."
                         .format(name))
    _policy_registry[sub_class][name] = policy



import tensorflow as tf
import numpy as np
import mpi4py

import shared.tf_util as tf_utils


class MpiAdam(object):
    def __init__(self, var_list, *, beta1=0.9, beta2=0.999, epsilon=1e-08, scale_grad_by_procs=True, comm=None,
                 sess=None):
        """
        A parallel MPI implementation of the Adam optimizer for TensorFlow
        https://arxiv.org/abs/1412.6980

        :param var_list: ([TensorFlow Tensor]) the variables
        :param beta1: (float) Adam beta1 parameter
        :param beta2: (float) Adam beta1 parameter
        :param epsilon: (float) to help with preventing arithmetic issues
        :param scale_grad_by_procs: (bool) if the scaling should be done by processes
        :param comm: (MPI Communicators) if None, mpi4py.MPI.COMM_WORLD
        :param sess: (TensorFlow Session) if None, tf.get_default_session()
        """
        self.var_list = var_list
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.scale_grad_by_procs = scale_grad_by_procs
        size = sum(tf_utils.numel(v) for v in var_list)
        # Exponential moving average of gradient values
        # "first moment estimate" m in the paper
        self.exp_avg = np.zeros(size, 'float32')
        # Exponential moving average of squared gradient values
        # "second raw moment estimate" v in the paper
        self.exp_avg_sq = np.zeros(size, 'float32')
        self.step = 0
        self.setfromflat = tf_utils.SetFromFlat(var_list, sess=sess)
        self.getflat = tf_utils.GetFlat(var_list, sess=sess)
        self.comm = mpi4py.MPI.COMM_WORLD if comm is None else comm

    def update(self, local_grad, learning_rate):
        """
        update the values of the graph

        :param local_grad: (numpy float) the gradient
        :param learning_rate: (float) the learning_rate for the update
        """
        if self.step % 100 == 0:
            self.check_synced()
        local_grad = local_grad.astype('float32')
        global_grad = np.zeros_like(local_grad)
        self.comm.Allreduce(local_grad, global_grad, op=mpi4py.MPI.SUM)
        if self.scale_grad_by_procs:
            global_grad /= self.comm.Get_size()

        self.step += 1
        # Learning rate with bias correction
        step_size = learning_rate * np.sqrt(1 - self.beta2 ** self.step) / (1 - self.beta1 ** self.step)
        # Decay the first and second moment running average coefficient
        self.exp_avg = self.beta1 * self.exp_avg + (1 - self.beta1) * global_grad
        self.exp_avg_sq = self.beta2 * self.exp_avg_sq + (1 - self.beta2) * (global_grad * global_grad)
        step = (- step_size) * self.exp_avg / (np.sqrt(self.exp_avg_sq) + self.epsilon)
        self.setfromflat(self.getflat() + step)

    def sync(self):
        """
        syncronize the MPI threads
        """
        theta = self.getflat()
        self.comm.Bcast(theta, root=0)
        self.setfromflat(theta)

    def check_synced(self):
        """
        confirm the MPI threads are synced
        """
        if self.comm.Get_rank() == 0:  # this is root
            theta = self.getflat()
            self.comm.Bcast(theta, root=0)
        else:
            thetalocal = self.getflat()
            thetaroot = np.empty_like(thetalocal)
            self.comm.Bcast(thetaroot, root=0)
            assert (thetaroot == thetalocal).all(), (thetaroot, thetalocal)


@tf_utils.in_session
def test_mpi_adam():
    """
    tests the MpiAdam object's functionality
    """
    np.random.seed(0)
    tf.set_random_seed(0)

    a_var = tf.Variable(np.random.randn(3).astype('float32'))
    b_var = tf.Variable(np.random.randn(2, 5).astype('float32'))
    loss = tf.reduce_sum(tf.square(a_var)) + tf.reduce_sum(tf.sin(b_var))

    learning_rate = 1e-2
    update_op = tf.train.AdamOptimizer(learning_rate).minimize(loss)
    do_update = tf_utils.function([], loss, updates=[update_op])

    tf.get_default_session().run(tf.global_variables_initializer())
    for step in range(10):
        print(step, do_update())

    tf.set_random_seed(0)
    tf.get_default_session().run(tf.global_variables_initializer())

    var_list = [a_var, b_var]
    lossandgrad = tf_utils.function([], [loss, tf_utils.flatgrad(loss, var_list)], updates=[update_op])
    adam = MpiAdam(var_list)

    for step in range(10):
        loss, grad = lossandgrad()
        adam.update(grad, learning_rate)
        print(step, loss)


if __name__ == "__main__":
    # Run with mpirun -np 2 python <filename>
    test_mpi_adam()



from typing import Union
from copy import deepcopy

import gym

# flake8: noqa F401
from neorl.rl.baselines.shared.vec_env.base_vec_env import AlreadySteppingError, NotSteppingError, VecEnv, VecEnvWrapper, \
    CloudpickleWrapper
from neorl.rl.baselines.shared.vec_env.dummy_vec_env import DummyVecEnv
from neorl.rl.baselines.shared.vec_env.subproc_vec_env import SubprocVecEnv
from neorl.rl.baselines.shared.vec_env.vec_frame_stack import VecFrameStack
from neorl.rl.baselines.shared.vec_env.vec_normalize import VecNormalize
from neorl.rl.baselines.shared.vec_env.vec_video_recorder import VecVideoRecorder
from neorl.rl.baselines.shared.vec_env.vec_check_nan import VecCheckNan


def unwrap_vec_normalize(env: Union[gym.Env, VecEnv]) -> Union[VecNormalize, None]:
    """
    :param env: (Union[gym.Env, VecEnv])
    :return: (VecNormalize)
    """
    env_tmp = env
    while isinstance(env_tmp, VecEnvWrapper):
        if isinstance(env_tmp, VecNormalize):
            return env_tmp
        env_tmp = env_tmp.venv
    return None


# Define here to avoid circular import
def sync_envs_normalization(env: Union[gym.Env, VecEnv], eval_env: Union[gym.Env, VecEnv]) -> None:
    """
    Sync eval and train environments when using VecNormalize

    :param env: (Union[gym.Env, VecEnv]))
    :param eval_env: (Union[gym.Env, VecEnv]))
    """
    env_tmp, eval_env_tmp = env, eval_env
    # Special case for the _UnvecWrapper
    # Avoid circular import
    from neorl.rl.baselines.shared.base_class import _UnvecWrapper
    if isinstance(env_tmp, _UnvecWrapper):
        return
    while isinstance(env_tmp, VecEnvWrapper):
        if isinstance(env_tmp, VecNormalize):
            # sync reward and observation scaling
            eval_env_tmp.obs_rms = deepcopy(env_tmp.obs_rms)
            eval_env_tmp.ret_rms = deepcopy(env_tmp.ret_rms)
        env_tmp = env_tmp.venv
        # Make pytype happy, in theory env and eval_env have the same type
        assert isinstance(eval_env_tmp, VecEnvWrapper), "the second env differs from the first env"
        eval_env_tmp = eval_env_tmp.venv



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
"""
Created on Thu Jul 29 09:57:06 2021

@author: Katelin Du
"""

import numpy as np
import matplotlib.pyplot as plt
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.callbacks import ModelCheckpoint
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler
import os

class NNmodel(object):
    def __init__(self, nn_params, gen, model_num, logger_paths=None):

        self.test_split = nn_params['test_split'] if 'test_split' in nn_params else 0.2
        self.activation = nn_params['activation'] if 'activation' in nn_params else 'relu'
        self.num_nodes = nn_params['num_nodes'] if 'num_nodes' in nn_params else [100, 50, 25]
        self.batch_size = nn_params['batch_size'] if 'batch_size' in nn_params else 32
        self.learning_rate = nn_params['learning_rate'] if 'learning_rate' in nn_params else 6e-4
        self.epochs = nn_params['epochs'] if 'epochs' in nn_params else 100
        self.plot_flag = nn_params['plot'] if 'plot' in nn_params else True
        self.verbose = nn_params['verbose'] if 'verbose' in nn_params else True
        self.save_models = nn_params['save_models'] if 'save_models' in nn_params else True

        self.gen = gen
        self.model_num = model_num
        self.paths=logger_paths
            
    def fit(self, X, Y):
        # """
        # Main function - generates model using NN parameters and saves the best of model of this generation.
        #
        # Return:
        # tuple - best model, maximum relative error from test set
        # """
        self.Xtrain, self.Xtest, self.Ytrain, self.Ytest = train_test_split(X, Y, test_size=self.test_split)
        # print(self.Xtrain.shape, self.Xtest.shape)
        xscaler = StandardScaler()     #x-scaler object
        yscaler = MinMaxScaler()     #y-scaler object
        self.Xtrain = xscaler.fit_transform(self.Xtrain)
        self.Xtest = xscaler.transform(self.Xtest)

        self.Ytrain = yscaler.fit_transform(self.Ytrain.reshape(-1,1)).flatten()
        self.Ytest = yscaler.transform(self.Ytest.reshape(-1,1)).flatten()

        model = self.model_structure()
        
        if self.save_models:
            cb = [ModelCheckpoint(filepath=os.path.join(self.paths['models'], 'model{0:0}_{1:04}.h5'.format(self.model_num, self.gen)), verbose=0, monitor='val_mean_absolute_error', save_best_only=True, mode='min')]
        else:
            cb = []

        if self.gen == 0:  #force saving the first the three models
            cb = [ModelCheckpoint(filepath='model{0:0}_{1:04}.h5'.format(self.model_num, self.gen), verbose=0, monitor='val_mean_absolute_error', save_best_only=True, mode='min')]

        model.compile(loss='mean_absolute_error', optimizer=Adam(self.learning_rate), metrics = ['mean_absolute_error'])

        #mir-new: consider these hyperparam: epochs, batch_size, validation_split
        self.history = model.fit(self.Xtrain, self.Ytrain, epochs=self.epochs, batch_size=self.batch_size, validation_data=(self.Xtest, self.Ytest), callbacks=cb, verbose=0)
        self.Ynn = model.predict(self.Xtest)
        self.Ynn = yscaler.inverse_transform(self.Ynn).flatten()
        self.Ytest = yscaler.inverse_transform(self.Ytest.reshape(-1,1)).flatten()
        # calculate relative error
        rel_errors = self.rel_errors()
        max_error = rel_errors.max()
        mean_error = rel_errors.mean()
        val_err=self.history.history['val_mean_absolute_error']

        if self.verbose:
            self.arg_max=np.argmax(rel_errors)
            print('-----------------------------------------------------')
            print(f'Generation {self.gen}, Model {self.model_num}')
            print('MAX relative test error:', max_error)
            print('MEAN relative test error:', mean_error)
            print('Val error:', val_err[-1])
            print('Worst target =', self.Ytest[self.arg_max])   #identify worst point
            print('Worst prediction =', self.Ynn[self.arg_max]) #identify worst point
            print()

        # graph prediction vs. target
        if self.plot_flag:
            self.plot()
            #print('Plots generated')
        return model

    def model_structure(self):
        # """
        # Defines the general structure of the neural network according to nn parameters.
        #
        # Return:
        # model structure
        # """
        model = Sequential()

        #mir-new: consider these hyperparam: num_nodes in list form (you infer num_dense_layers)
        num_dense_layers = len(self.num_nodes)
        #mir-new: consider activation in HIDDEN Layers ONLY (relu, sigmoid, etc.)
        model.add(Dense(self.num_nodes[0], kernel_initializer='normal', activation=self.activation, input_dim=self.Xtrain.shape[1]))
        model.add(Dropout(0.5))
        for i in range(1, num_dense_layers):
            model.add(Dense(self.num_nodes[i], activation=self.activation, kernel_initializer='normal'))
        model.add(Dense(1, activation='linear', kernel_initializer='normal'))
        return model

    def rel_errors(self):
        # """
        # Finds relative errors between target test set and predicted set.
        #
        # Return:
        # array - relative errors
        # """
        # abs error / true value
        rel_errors = []
        for i in range(len(self.Ynn)):
            rel_errors.append(100 * abs(self.Ytest[i] - self.Ynn[i]) / self.Ytest[i])
        return np.array(rel_errors)

    def plot(self):
        # """
        # Saves relevant plots for the current generation:
        # 1. Target vs. Predicted
        # 2. Loss (MAE) vs. Epoch
        # """
        plt.figure()
        plt.plot(self.Ytest, self.Ynn, 'o')
        max_val = max([max(self.Ytest), max(self.Ynn)])
        plt.plot(range(0,int(max_val)), range(0,int(max_val)),'-k')
        plt.xlabel('Target')
        plt.ylabel('Predicted')
        plt.savefig(os.path.join(self.paths['predict'], 'prediction{0:0}_{1:04}.png'.format(self.model_num, self.gen))) #mir: make the name as prediction-00001.png, .... based on generation #
        plt.close()

        # obtain the training/validation MAE from fitting history
        train_err=self.history.history['mean_absolute_error']
        val_err=self.history.history['val_mean_absolute_error']
        plt.figure()
        plt.plot(train_err, label='Training')
        plt.plot(val_err, label='Validation')
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.legend()
        plt.savefig(os.path.join(self.paths['error'], 'error{0:0}_{1:04}.png'.format(self.model_num, self.gen)))  #mir: same as the previous figure
        plt.close()



#!/usr/bin/env python3
import numpy as np
import gym

from neorl.rl.baselines.shared.cmd_util import mujoco_arg_parser
from neorl.rl.baselines.shared import logger
from neorl.rl.baselines.shared.monitor import Monitor
from neorl.rl.baselines.shared import set_global_seeds
from neorl.rl.baselines.shared.vec_env.vec_normalize import VecNormalize
from neorl.rl.baselines.ppo2.ppo2 import PPO2
from neorl.rl.baselines.shared.policies import MlpPolicy
from neorl.rl.baselines.shared.vec_env.dummy_vec_env import DummyVecEnv


def train(env_id, num_timesteps, seed):
    """
    Train PPO2 model for Mujoco environment, for testing purposes

    :param env_id: (str) the environment id string
    :param num_timesteps: (int) the number of timesteps to run
    :param seed: (int) Used to seed the random generator.
    """
    def make_env():
        env_out = gym.make(env_id)
        env_out = Monitor(env_out, logger.get_dir(), allow_early_resets=True)
        return env_out

    env = DummyVecEnv([make_env])
    env = VecNormalize(env)

    set_global_seeds(seed)
    policy = MlpPolicy
    model = PPO2(policy=policy, env=env, n_steps=2048, nminibatches=32, lam=0.95, gamma=0.99, noptepochs=10,
                 ent_coef=0.0, learning_rate=3e-4, cliprange=0.2)
    model.learn(total_timesteps=num_timesteps)

    return model, env


def main():
    """
    Runs the test
    """
    args = mujoco_arg_parser().parse_args()
    logger.configure()
    model, env = train(args.env, num_timesteps=args.num_timesteps, seed=args.seed)

    if args.play:
        logger.log("Running trained model")
        obs = np.zeros((env.num_envs,) + env.observation_space.shape)
        obs[:] = env.reset()
        while True:
            actions = model.step(obs)[0]
            obs[:] = env.step(actions)[0]
            env.render()


if __name__ == '__main__':
    main()



import pandas as pd
from scipy.interpolate import interp1d
import sys
import inspect
from pathlib import Path
import numpy as np

cpath = Path(inspect.getfile(sys.modules[__name__])).resolve().parent

def get_jminus(typ = "wtd"): #wtd, refl or abs
    """Retrieve j- function from model data files."""
    #pull and load data
    configA_path = cpath / Path("tools/microreactor_configA_flux_%s_b.csv"%typ)
    configB_path = cpath / Path("tools/microreactor_configB_flux_%s_b.csv"%typ)

    jm_configA = pd.read_csv(configA_path, index_col = 0)
    jm_configB = pd.read_csv(configB_path, index_col = 0)

    #adjust 0 angle to be inward to the reactor
    jm_configA["centers"] -= 3.6820
    jm_configB["centers"] -= 4.0677

    #center periodic functions on 0
    jm_configA["centers"][jm_configA["centers"] < -np.pi] += 2*np.pi
    jm_configB["centers"][jm_configB["centers"] < -np.pi] += 2*np.pi

    #add buffer on each side for periodic integral evaluations
    lowcopy = jm_configA.copy()
    lowcopy["centers"] -= 2*np.pi
    highcopy = jm_configA.copy()
    highcopy["centers"] += 2*np.pi
    jm_configA = pd.concat([jm_configA, lowcopy, highcopy])

    lowcopy = jm_configB.copy()
    lowcopy["centers"] -= 2*np.pi
    highcopy = jm_configB.copy()
    highcopy["centers"] += 2*np.pi
    jm_configB = pd.concat([jm_configB, lowcopy, highcopy])

    #reorder from -pi to pi
    jm_configA = jm_configA.sort_values(by="centers")
    jm_configB = jm_configB.sort_values(by="centers")

    return jm_configA, jm_configB

def get_alphas(typ = "wtd"): #wtd, refl, abs
    """read alphas from file"""
    fpath = cpath / Path("tools/microreactor_alpha_%s_gr.csv"%typ)
    return pd.read_csv(fpath, index_col = 0)

def adj_coords(thetas):
    """brings array of angles into [-np.pi, np.pi]"""
    thetas = thetas % (2*np.pi)
    thetas[thetas > np.pi] -= 2*np.pi
    return thetas

def integrate(x, y, lbnd, ubnd):
    """Integrate j- function given x & y across the bounds."""
    #make sure dealing with numpy arrays
    x, y = np.asarray(x), np.asarray(y)

    #identify portions of functions where x-blocks are completely enclosed
    compl_ind = np.where(np.logical_and(x > lbnd, x < ubnd))[0]

    if compl_ind.size == 0: #if both bounds fall within an interval
        idx = np.searchsorted(x, lbnd)
        y_lbnd_approx = (lbnd - x[idx-1])/(x[idx] - x[idx-1])*(y[idx] - y[idx-1]) + y[idx-1]
        y_ubnd_approx = (ubnd - x[idx-1])/(x[idx] - x[idx-1])*(y[idx] - y[idx-1]) + y[idx-1]
        return (ubnd - lbnd)*(y_ubnd_approx + y_lbnd_approx)/2

    #integrate fully-inclosed blocks
    full_blocks_integral = np.trapz(y[compl_ind], x[compl_ind])

    #integrate lower hanging partial block
    lidx = compl_ind.min()
    y_lbnd_approx = (lbnd - x[lidx-1])/(x[lidx] - x[lidx-1])*(y[lidx] - y[lidx-1]) + y[lidx-1]
    lower_block_integral = (x[lidx] - lbnd)*(y[lidx] + y_lbnd_approx)/2

    #integrate upper hanging partial block
    uidx = compl_ind.max()+1
    y_ubnd_approx = (ubnd - x[uidx-1])/(x[uidx] - x[uidx-1])*(y[uidx] - y[uidx-1]) + y[uidx-1]
    upper_block_integral = (ubnd - x[uidx - 1])*(y[uidx - 1] + y_ubnd_approx)/2

    return full_blocks_integral + lower_block_integral + upper_block_integral

def integrate_sq(x, y, lbnd, ubnd):
    """Integrate j-^2 function given x & y across the bounds. y will be squared within the routine"""
    #make sure dealing with numpy arrays
    x, y = np.asarray(x), np.asarray(y)

    #identify portions of functions where x-blocks are completely enclosed
    compl_ind = np.where(np.logical_and(x > lbnd, x < ubnd))[0]

    if compl_ind.size == 0: #if both bounds fall within an interval
        idx = np.searchsorted(x, lbnd)
        y_lbnd_approx = (lbnd - x[idx-1])/(x[idx] - x[idx-1])*(y[idx] - y[idx-1]) + y[idx-1]
        y_ubnd_approx = (ubnd - x[idx-1])/(x[idx] - x[idx-1])*(y[idx] - y[idx-1]) + y[idx-1]
        m = (y_ubnd_approx - y_lbnd_approx)/(ubnd - lbnd)
        return m**2/3*(ubnd**3 - lbnd**3) + (m*y_lbnd_approx - m**2*lbnd)*(ubnd**2 - lbnd**2)\
                + (m*lbnd - y_lbnd_approx)**2*(ubnd - lbnd)

    #integrate fully-inclosed blocks
    dx = np.diff(x[compl_ind]) #xi+1 - xi
    dx2 = np.diff(x[compl_ind]**2) #xi+1**2 - xi**2
    dx3 = np.diff(x[compl_ind]**3) #xi+1**3 - xi**3
    ms = np.diff(y[compl_ind])/dx
    xis = x[compl_ind[:-1]]
    yis = y[compl_ind[:-1]]
    full_blocks_integral = (ms**2*dx3/3+(ms*yis-ms**2*xis)*dx2+(ms*xis-yis)**2*dx).sum()

    #integrate lower hanging partial block
    lidx = compl_ind.min()
    y_lbnd_approx = (lbnd - x[lidx-1])/(x[lidx] - x[lidx-1])*(y[lidx] - y[lidx-1]) + y[lidx-1]
    m = (y[lidx] - y_lbnd_approx)/(x[lidx] - lbnd)
    lower_block_integral = m**2/3*(x[lidx]**3 - lbnd**3) \
            + (m*y_lbnd_approx-m**2*lbnd)*(x[lidx]**2 - lbnd**2) \
            + (m*lbnd - y_lbnd_approx)**2*(x[lidx] - lbnd)

    #integrate upper hanging partial block
    uidx = compl_ind.max()
    y_ubnd_approx = (ubnd - x[uidx])/(x[uidx+1] - x[uidx])*(y[uidx+1] - y[uidx]) + y[uidx]
    m = (y_ubnd_approx - y[uidx])/(ubnd - x[uidx])
    upper_block_integral = m**2/3*(ubnd**3 - x[uidx]**3) \
            + (m*y[uidx]-m**2*x[uidx])*(ubnd**2 - x[uidx]**2) \
            + (m*x[uidx] - y[uidx])**2*(ubnd - x[uidx])

    return full_blocks_integral + lower_block_integral + upper_block_integral

def int_bounds(theta, cangle):
    """get bounds on j-^2 integrals given rotation angle, theta and coating angle cangle
    returns (thetaA&thetaA0'), (thetaA'&thetaA0)"""
    assert (theta < np.pi + 1e-5) and (theta > -np.pi - 1e-5)

    if 0 < theta and theta < cangle:
        return ([cangle/2, theta + cangle/2], [-cangle/2, theta - cangle/2])
    elif -cangle < theta and theta < 0:
        return ([theta - cangle/2, -cangle/2], [theta + cangle/2, cangle/2])
    else:
        return ([theta - cangle/2, theta + cangle/2], [-cangle/2, cangle/2])

def calc_zetatildes(theta, cangles, alphas, jminusA, jminusB):
    """
    calculate zetatilde functions for all drums
    thetas is numpy array of size 8 of rotation angles
    cangles is numpy array of size 8 of coating angles
    alphas is numpy array of shape (8,9) of model coefficients
    jminusA is pandas array with "centers" and "hist" columns
    jminusB is pandas array with "centers" and "hist" columns
    """
    configAids = [1, 4, 5, 8] #drum positions with configuration A

    #calculate gammastar by integrting j- over abs bounds for each drum
    gammastar = np.ones(9)
    for i in range(1, 9):
        if i in configAids:
            gammastar[i] = 1 - integrate(x = jminusA["centers"],
                                         y = jminusA["hist"],
                                         lbnd = theta[i-1] - cangles[i-1]/2,
                                         ubnd = theta[i-1] + cangles[i-1]/2)
        else:
            gammastar[i] = 1 - integrate(x = jminusB["centers"],
                                         y = jminusB["hist"],
                                         lbnd = theta[i-1] - cangles[i-1]/2,
                                         ubnd = theta[i-1] + cangles[i-1]/2)

    #calculate zetatilde for each drum
    return (gammastar@alphas.T).values

def calc_dzetatildes(thetas, k, cangles, alphas, jmfs):
    """
    calculate derivative zetatilde functions for all drums resct to drum k
    thetas is numpy array of size 8 of rotation angles
    k is which of drums angle is taken respect to (1-indexed)
    cangles is numpy array of size 8 of coating angles
    alphas is numpy array of shape (8,9) of model coefficients
    jmfs is list of functions for inward current
    return is 8 element array
    """
    alpha_slice = alphas["alpha" + str(k)].values
    dgamma = -jmfs[k-1](thetas[k-1] + cangles[k-1]/2) + jmfs[k-1](thetas[k-1] - cangles[k-1]/2)
    return alpha_slice*dgamma

def calc_W(theta, cangle, jmf):
    """
    Calculate derivative of difference of integrals
    theta is k-th drum angle
    cangles is k-th drum coating angle
    jmf is jminus of k-th rum
    """
    t1 = jmf(theta + cangle/2)**2
    t2 = jmf(theta - cangle/2)**2
    return t1 - t2

class ReactivityModel:
    """
    Used to evaluate reactivity insertion from control drum perturbation.
    Set up as init->method call to minimize file reading times
    """
    def __init__(self, typ = "wtd"): #abs, wtd or refl
        """initialize to perform all file I/O"""
        self.jmA, self.jmB = get_jminus(typ)
        self.alphas = get_alphas(typ)
        self.cangles = np.array([130, 145, 145, 130,
                                 130, 145, 145, 130])/180*np.pi
        fA = interp1d(self.jmA["centers"].values, self.jmA["hist"].values)
        fB = interp1d(self.jmB["centers"].values, self.jmB["hist"].values)
        self.jmfs = [fA, fB, fB, fA, fA, fB, fB, fA]

    def eval(self, pert, nom = None):
        """
        Evaluate reactivity worth of drum perturbation.
        Pert is numpy array of 8 drum angles in radians with 
        coordinate systems described in the README.md.
        Nom is an optional starting state given same as pert
        """
        #bring drum angles into [-np.pi, np.pi]
        pert = adj_coords(pert)

        #pull in zetatildes from external function
        zetatildes = calc_zetatildes(theta = pert,
                                     cangles = self.cangles,
                                     alphas = self.alphas,
                                     jminusA = self.jmA,
                                     jminusB = self.jmB)

        #loop through drums and calculate each contribution
        reactivities = np.zeros(8)
        for i in range(8):
            b1, b2 = int_bounds(pert[i], self.cangles[i])
            if i in [0, 3, 4, 7]: #if config A
                int1 = integrate_sq(self.jmA["centers"], self.jmA["hist"], *b1)
                int2 = integrate_sq(self.jmA["centers"], self.jmA["hist"], *b2)
            else: #if config A
                int1 = integrate_sq(self.jmB["centers"], self.jmB["hist"], *b1)
                int2 = integrate_sq(self.jmB["centers"], self.jmB["hist"], *b2)
            reactivities[i] = zetatildes[i]*(int1 - int2)

        #basically just return conditionals
        if nom: #little trick
            return self.eval(pert) - self.eval(nom) #assume
                                                          #reactivites additive
        else:
            return reactivities.sum()

    def evald(self, pert, k, zetatildes = None):
        """
        Evaluate differential reactivity worth of drum config from single drum.
        Pert is numpy array of 8 drum angles in radians with 
        k is which drum rotation to take derivative respect to
        zetatildes should be ignored, used for computational efficiency
        coordinate systems described in the README.md.
        """
        #bring drum angles into [-np.pi, np.pi]
        pert = adj_coords(pert)

        #calculate zetatilde parameters
        dzetatildes = calc_dzetatildes(thetas = pert,
                                      k = k,
                                      cangles = self.cangles,
                                      alphas = self.alphas,
                                      jmfs = self.jmfs)
        if zetatildes is None:
            zetatildes = calc_zetatildes(theta = pert,
                                         cangles = self.cangles,
                                         alphas = self.alphas,
                                         jminusA = self.jmA,
                                         jminusB = self.jmB)

        #define vector to hold summation
        drdtk = np.zeros(8)

        #iterate through all drums
        for i in range(8):
            b1, b2 = int_bounds(pert[i], self.cangles[i])
            if i in [0, 3, 4, 7]: #if config A
                int1 = integrate_sq(self.jmA["centers"], self.jmA["hist"], *b1)
                int2 = integrate_sq(self.jmA["centers"], self.jmA["hist"], *b2)
            else: #if config B
                int1 = integrate_sq(self.jmB["centers"], self.jmB["hist"], *b1)
                int2 = integrate_sq(self.jmB["centers"], self.jmB["hist"], *b2)

            drdtk[i] += dzetatildes[i] * (int1 - int2)

        #tack on extra term from eq41
        drdtk[k-1] += zetatildes[k-1]*calc_W(pert[k-1], self.cangles[k-1], self.jmfs[k-1])

        return drdtk.sum()

    def evalg(self, pert):
        """
        Evaluate differential reactivity worth of drum config for all drums.
        Equivalent to the gradient.
        Pert is numpy array of 8 drum angles in radians with 
        coordinate systems described in the README.md.
        """
        zetatildes = calc_zetatildes(theta = pert,
                                     cangles = self.cangles,
                                     alphas = self.alphas,
                                     jminusA = self.jmA,
                                     jminusB = self.jmB)
        grad = np.zeros(8)
        for i in range(8):
            grad[i] = self.evald(pert, i+1, zetatildes)
        return grad

def reactivityModelEval(pert, nom = None, typ = "wtd"):
    """Wrapper for ReactivityModel that initializes and runs"""
    a = ReactivityModel(typ)
    return a.eval(pert, nom)

def reactivityModelEvald(pert, k, typ = "wtd"):
    """Wrapper for ReactivityModel that initializes and runs"""
    aa = ReactivityModel(typ)
    return aa.evald(pert, k)

def reactivityModelEvalg(pert, typ = "wtd"):
    """Wrapper for ReactivityModel that initializes and runs"""
    aa = ReactivityModel(typ)
    return aa.evalg(pert)

if __name__ == "__main__":
    a = ReactivityModel()
    p5 = np.array([0, 0, 2.201331,-2.137274,-2.512146,2.02961,0.19566370000000002,3.140272])
    print(a.eval(p5))
